from __future__ import annotations

import abc
import base64
import io
import itertools
import pickle
import typing as t
from typing import Any

from ..types import LazyType
from ..utils.lazy_loader import LazyLoader
from ..utils.pickle import fixed_torch_loads
from ..utils.pickle import pep574_dumps
from ..utils.pickle import pep574_loads
from .utils import Params

SingleType = t.TypeVar("SingleType")
BatchType = t.TypeVar("BatchType")
T = t.TypeVar("T")

TRITON_EXC_MSG = "tritonclient is required to use triton with BentoML. Install with 'pip install \"tritonclient[all]>=2.29.0\"'."

if t.TYPE_CHECKING:
    import numpy as np
    import tritonclient.grpc.aio as tritongrpcclient
    import tritonclient.http.aio as tritonhttpclient

    from .. import external_typing as ext

else:
    np = LazyLoader("np", globals(), "numpy")
    tritongrpcclient = LazyLoader(
        "tritongrpcclient", globals(), "tritonclient.grpc.aio", exc_msg=TRITON_EXC_MSG
    )
    tritonhttpclient = LazyLoader(
        "tritonhttpclient", globals(), "tritonclient.http.aio", exc_msg=TRITON_EXC_MSG
    )

InferInput = t.Union["tritongrpcclient.InferInput", "tritonhttpclient.InferInput"]


class Payload(t.NamedTuple):
    data: bytes
    meta: dict[str, bool | int | float | str | list[int]]
    container: str
    batch_size: int = -1


class DataContainer(t.Generic[SingleType, BatchType]):
    @classmethod
    def create_payload(
        cls,
        data: bytes,
        batch_size: int,
        meta: dict[str, bool | int | float | str | list[int]] | None = None,
    ) -> Payload:
        return Payload(data, meta or {}, container=cls.__name__, batch_size=batch_size)

    @classmethod
    @abc.abstractmethod
    def to_payload(cls, batch: BatchType, batch_dim: int) -> Payload: ...

    @classmethod
    @abc.abstractmethod
    def from_payload(cls, payload: Payload) -> BatchType: ...

    @classmethod
    def batch_to_payloads(
        cls, batch: BatchType, indices: t.Sequence[int], batch_dim: int = 0
    ) -> list[Payload]:
        batches = cls.batch_to_batches(batch, indices, batch_dim)
        payloads = [cls.to_payload(subbatch, batch_dim) for subbatch in batches]
        return payloads

    @classmethod
    def from_batch_payloads(
        cls, payloads: t.Sequence[Payload], batch_dim: int = 0
    ) -> tuple[BatchType, list[int]]:
        batches = [cls.from_payload(payload) for payload in payloads]
        return cls.batches_to_batch(batches, batch_dim)

    @t.overload
    @classmethod
    def to_triton_payload(
        cls,
        inp: tritongrpcclient.InferInput,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
        _use_http_client: t.Literal[False] = ...,
    ) -> tritongrpcclient.InferInput: ...

    @t.overload
    @classmethod
    def to_triton_payload(
        cls,
        inp: tritonhttpclient.InferInput,
        meta: dict[str, t.Any],
        _use_http_client: t.Literal[True] = ...,
    ) -> tritongrpcclient.InferInput: ...

    @classmethod
    def to_triton_payload(
        cls, inp: SingleType, meta: t.Any, _use_http_client: bool = False
    ) -> t.Any:
        """
        Convert given input types to a Triton payload.

        Implementation of this method should always return a InferInput that has
        data of a Numpy array.
        """
        if _use_http_client:
            return cls.to_triton_http_payload(inp, meta)
        else:
            return cls.to_triton_grpc_payload(inp, meta)

    @classmethod
    def to_triton_grpc_payload(
        cls,
        inp: SingleType,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
    ) -> tritongrpcclient.InferInput:
        """
        Convert given input types to a Triton payload via gRPC client.

        Implementation of this method should always return a InferInput.
        """
        raise NotImplementedError(
            f"{cls.__name__} doesn't support converting to Triton payload."
        )

    @classmethod
    def to_triton_http_payload(
        cls,
        inp: SingleType,
        meta: dict[str, t.Any],
    ) -> tritonhttpclient.InferInput:
        """
        Convert given input types to a Triton payload via HTTP client.

        Implementation of this method should always return a InferInput.
        """
        raise NotImplementedError(
            f"{cls.__name__} doesn't support converting to Triton payload."
        )

    @classmethod
    @abc.abstractmethod
    def batches_to_batch(
        cls, batches: t.Sequence[SingleType], batch_dim: int
    ) -> tuple[BatchType, list[int]]: ...

    @classmethod
    @abc.abstractmethod
    def batch_to_batches(
        cls, batch: BatchType, indices: t.Sequence[int], batch_dim: int
    ) -> list[SingleType]: ...

    @classmethod
    def get_batch_size(cls, batch: SingleType, batch_dim: int) -> int:
        """A default implementation for backward compatibility"""
        return 1


class TritonInferInputDataContainer(DataContainer[InferInput, InferInput]):
    @classmethod
    def to_payload(cls, batch: InferInput, batch_dim: int) -> Payload:
        raise NotImplementedError

    @classmethod
    def from_payload(cls, payload: Payload) -> InferInput:
        raise NotImplementedError

    @classmethod
    def to_triton_grpc_payload(
        cls,
        inp: InferInput,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
    ) -> tritongrpcclient.InferInput:
        return t.cast("tritongrpcclient.InferInput", inp)

    @classmethod
    def to_triton_http_payload(
        cls,
        inp: InferInput,
        meta: dict[str, t.Any],
    ) -> tritonhttpclient.InferInput:
        return t.cast("tritonhttpclient.InferInput", inp)

    @classmethod
    def batches_to_batch(
        cls,
        batches: t.Sequence[InferInput],
        batch_dim: int,
    ) -> tuple[InferInput, list[int]]:
        raise NotImplementedError

    @classmethod
    def batch_to_batches(
        cls,
        batch: InferInput,
        indices: t.Sequence[int],
        batch_dim: int,
    ) -> list[InferInput]:
        raise NotImplementedError


class NdarrayContainer(DataContainer["ext.NpNDArray", "ext.NpNDArray"]):
    @classmethod
    def batches_to_batch(
        cls,
        batches: t.Sequence[ext.NpNDArray],
        batch_dim: int = 0,
    ) -> tuple[ext.NpNDArray, list[int]]:
        # numpy.concatenate may consume lots of memory, need optimization later
        batch: ext.NpNDArray = np.concatenate(batches, axis=batch_dim)
        indices = list(
            itertools.accumulate(subbatch.shape[batch_dim] for subbatch in batches)
        )
        indices = [0] + indices
        return batch, indices

    @classmethod
    def batch_to_batches(
        cls,
        batch: ext.NpNDArray,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[ext.NpNDArray]:
        return np.split(batch, indices[1:-1], axis=batch_dim)

    @classmethod
    def to_triton_grpc_payload(
        cls,
        inp: ext.NpNDArray,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
    ) -> tritongrpcclient.InferInput:
        InferInput = tritongrpcclient.InferInput(
            meta.name, inp.shape, tritongrpcclient.np_to_triton_dtype(inp.dtype)
        )
        InferInput.set_data_from_numpy(inp)
        return InferInput

    @classmethod
    def to_triton_http_payload(
        cls,
        inp: ext.NpNDArray,
        meta: dict[str, t.Any],
    ) -> tritonhttpclient.InferInput:
        InferInput = tritonhttpclient.InferInput(
            meta["name"], inp.shape, tritonhttpclient.np_to_triton_dtype(inp.dtype)
        )
        InferInput.set_data_from_numpy(inp)
        return InferInput

    @classmethod
    def to_payload(
        cls,
        batch: ext.NpNDArray,
        batch_dim: int,
    ) -> Payload:
        # skip 0-dimensional array
        if batch.shape:
            if not (batch.flags["C_CONTIGUOUS"] or batch.flags["F_CONTIGUOUS"]):
                # TODO: use fortan contiguous if it's faster
                batch = np.ascontiguousarray(batch)

            meta: dict[str, bool | int | float | str | list[int]] = {
                "format": "pickle5"
            }

            bs: bytes
            concat_buffer_bs: bytes
            indices: list[int]
            bs, concat_buffer_bs, indices = pep574_dumps(batch)

            if indices:
                if len(bs) > 6140:  # ceil(6140 / 3) * 4 = 8188 < 8192
                    meta["format"] = "default"
                    data = pickle.dumps(batch)
                else:
                    meta["with_buffer"] = True
                    data = concat_buffer_bs
                    meta["pickle_bytes_str"] = base64.b64encode(bs).decode("ascii")
                    meta["indices"] = indices
            else:
                meta["with_buffer"] = False
                data = bs

            return cls.create_payload(
                data,
                batch.shape[batch_dim],
                meta=meta,
            )

        return cls.create_payload(
            pickle.dumps(batch),
            batch.shape[batch_dim],
            {"format": "default"},
        )

    @classmethod
    def get_batch_size(cls, batch: ext.NpNDArray, batch_dim: int) -> int:
        return batch.shape[batch_dim]

    @classmethod
    def from_payload(
        cls,
        payload: Payload,
    ) -> ext.NpNDArray:
        format = payload.meta.get("format", "default")
        if format == "default":
            return pickle.loads(payload.data)

        # format can be either "default" or "pickle5"
        if payload.meta["with_buffer"]:
            bs_str = t.cast(str, payload.meta["pickle_bytes_str"])
            bs = base64.b64decode(bs_str)
            indices = t.cast(t.List[int], payload.meta["indices"])
            return t.cast("ext.NpNDArray", pep574_loads(bs, payload.data, indices))
        else:
            return t.cast("ext.NpNDArray", pep574_loads(payload.data, b"", []))


class PandasDataFrameContainer(
    DataContainer[t.Union["ext.PdDataFrame", "ext.PdSeries"], "ext.PdDataFrame"]
):
    @classmethod
    def batches_to_batch(
        cls,
        batches: t.Sequence[ext.PdDataFrame],
        batch_dim: int = 0,
    ) -> tuple[ext.PdDataFrame, list[int]]:
        import pandas as pd

        assert batch_dim == 0, (
            "PandasDataFrameContainer does not support batch_dim other than 0"
        )
        indices = list(
            itertools.accumulate(subbatch.shape[batch_dim] for subbatch in batches)
        )
        indices = [0] + indices
        return pd.concat(batches, ignore_index=True), indices  # type: ignore (incomplete panadas types)

    @classmethod
    def batch_to_batches(
        cls,
        batch: ext.PdDataFrame | ext.PdSeries,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[ext.PdDataFrame]:
        assert batch_dim == 0, (
            "PandasDataFrameContainer does not support batch_dim other than 0"
        )

        return [
            batch.iloc[indices[i] : indices[i + 1]].reset_index(drop=True)
            for i in range(len(indices) - 1)
        ]

    @classmethod
    def to_payload(
        cls,
        batch: ext.PdDataFrame | ext.PdSeries,
        batch_dim: int,
    ) -> Payload:
        import pandas as pd

        assert batch_dim == 0, (
            "PandasDataFrameContainer does not support batch_dim other than 0"
        )

        if isinstance(batch, pd.Series):
            batch = pd.DataFrame([batch])

        meta: dict[str, bool | int | float | str | list[int]] = {"format": "pickle5"}

        bs: bytes
        concat_buffer_bs: bytes
        indices: list[int]
        bs, concat_buffer_bs, indices = pep574_dumps(batch)

        if indices:
            if len(bs) > 6140:  # ceil(6140 / 3) * 4 = 8188 < 8192
                meta["format"] = "default"
                data = pickle.dumps(batch)
            else:
                meta["with_buffer"] = True
                data = concat_buffer_bs
                meta["pickle_bytes_str"] = base64.b64encode(bs).decode("ascii")
                meta["indices"] = indices
        else:
            meta["with_buffer"] = False
            data = bs

        return cls.create_payload(
            data,
            batch.shape[0],
            meta=meta,
        )

    @classmethod
    def get_batch_size(
        cls, batch: ext.PdDataFrame | ext.PdSeries, batch_dim: int
    ) -> int:
        assert batch_dim == 0, (
            "PandasDataFrameContainer does not support batch_dim other than 0"
        )
        return batch.shape

    @classmethod
    def from_payload(
        cls,
        payload: Payload,
    ) -> ext.PdDataFrame:
        if payload.meta["format"] == "default":
            return pickle.loads(payload.data)

        if payload.meta["with_buffer"]:
            bs_str = t.cast(str, payload.meta["pickle_bytes_str"])
            bs = base64.b64decode(bs_str)
            indices = t.cast(t.List[int], payload.meta["indices"])
            return pep574_loads(bs, payload.data, indices)
        else:
            return pep574_loads(payload.data, b"", [])

    @classmethod
    def from_batch_payloads(  # pylint: disable=arguments-differ
        cls,
        payloads: t.Sequence[Payload],
        batch_dim: int = 0,
    ) -> tuple[ext.PdDataFrame, list[int]]:
        batches = [cls.from_payload(payload) for payload in payloads]
        return cls.batches_to_batch(batches, batch_dim)


class PILImageContainer(DataContainer["ext.PILImage", "ext.PILImage"]):
    _error = (
        "PIL.Image doesn't support batch inference."
        "You can convert it to numpy.ndarray before passing to the runner."
    )

    @classmethod
    def to_payload(cls, batch: ext.PILImage, batch_dim: int) -> Payload:
        buffer = io.BytesIO()
        batch.save(buffer, format=batch.format)
        return cls.create_payload(buffer.getvalue(), batch_size=1)

    @classmethod
    def from_payload(cls, payload: Payload) -> ext.PILImage:
        from ..io_descriptors.image import PIL

        return PIL.Image.open(io.BytesIO(payload.data))

    @classmethod
    def batches_to_batch(
        cls, batches: t.Sequence[ext.PILImage], batch_dim: int
    ) -> tuple[ext.PILImage, list[int]]:
        raise NotImplementedError(cls._error)

    @classmethod
    def batch_to_batches(
        cls, batch: ext.PILImage, indices: t.Sequence[int], batch_dim: int
    ) -> list[ext.PILImage]:
        raise NotImplementedError(cls._error)


class ParamsContainer(DataContainer[Params[Payload], Params[Payload]]):
    @classmethod
    def batches_to_batch(
        cls, batches: t.Sequence[Params[Payload]], batch_dim: int
    ) -> tuple[Params[Payload], list[int]]:
        converted = Params.agg(
            batches,
            agg_func=lambda payloads: AutoContainer.from_batch_payloads(
                payloads, batch_dim
            ),
        )
        batch_param, indices, *_ = converted.iter()
        return t.cast("Params[t.Any]", batch_param).map(
            lambda x: AutoContainer.to_payload(x, batch_dim)
        ), t.cast("Params[list[int]]", indices).sample

    @classmethod
    def batch_to_batches(
        cls, batch: Params[Payload], indices: t.Sequence[int], batch_dim: int
    ) -> list[Params[Payload]]:
        def payload_to_batches(payload: Payload) -> list[Payload]:
            container_cls = DataContainerRegistry.find_by_name(payload.container)
            batches = container_cls.batch_to_batches(
                container_cls.from_payload(payload), indices, batch_dim
            )
            return [
                container_cls.to_payload(subbatch, batch_dim) for subbatch in batches
            ]

        batches_params = batch.map(payload_to_batches)
        return list(batches_params.iter())

    @classmethod
    def to_payload(cls, batch: Params[Payload], batch_dim: int) -> Payload:
        raise NotImplementedError

    @classmethod
    def from_payload(cls, payload: Payload) -> Params[Payload]:
        raise NotImplementedError

    @classmethod
    def get_batch_size(cls, batch: Params[Payload], batch_dim: int) -> int:
        return batch.sample.batch_size


class PayloadContainer(DataContainer[Payload, Payload]):
    @classmethod
    def to_payload(cls, batch: Payload, batch_dim: int) -> Payload:
        raise NotImplementedError

    @classmethod
    def from_payload(cls, payload: Payload) -> Payload:
        raise NotImplementedError

    @classmethod
    def batch_to_batches(
        cls, batch: Payload, indices: t.Sequence[int], batch_dim: int
    ) -> list[Payload]:
        container_cls = DataContainerRegistry.find_by_name(batch.container)
        return container_cls.batch_to_payloads(
            container_cls.from_payload(batch), indices, batch_dim
        )

    @classmethod
    def batches_to_batch(
        cls, batches: t.Sequence[Payload], batch_dim: int
    ) -> tuple[Payload, list[int]]:
        container_cls = DataContainerRegistry.find_by_name(batches[0].container)
        data, indices = container_cls.from_batch_payloads(batches, batch_dim)
        return container_cls.to_payload(data, batch_dim), indices

    @classmethod
    def get_batch_size(cls, batch: Payload, batch_dim: int) -> int:
        return batch.batch_size


class DefaultContainer(DataContainer[t.Any, t.List[t.Any]]):
    @classmethod
    def batches_to_batch(
        cls, batches: t.Sequence[list[t.Any]], batch_dim: int = 0
    ) -> tuple[list[t.Any], list[int]]:
        assert batch_dim == 0, (
            "Default Runner DataContainer does not support batch_dim other than 0"
        )
        batch: list[t.Any] = []
        for subbatch in batches:
            batch.extend(subbatch)
        indices = list(itertools.accumulate(len(subbatch) for subbatch in batches))
        indices = [0] + indices
        return batch, indices

    @classmethod
    def batch_to_batches(
        cls, batch: list[t.Any], indices: t.Sequence[int], batch_dim: int = 0
    ) -> list[list[t.Any]]:
        assert batch_dim == 0, (
            "Default Runner DataContainer does not support batch_dim other than 0"
        )
        return [batch[indices[i] : indices[i + 1]] for i in range(len(indices) - 1)]

    @classmethod
    def to_payload(cls, batch: t.Any, batch_dim: int) -> Payload:
        if isinstance(batch, t.Generator):  # Generators can't be pickled
            batch = list(t.cast(t.Generator[t.Any, t.Any, t.Any], batch))

        data = pickle.dumps(batch)

        if isinstance(batch, list):
            batch_size = len(t.cast(t.List[t.Any], batch))
        else:
            batch_size = 1

        return cls.create_payload(data=data, batch_size=batch_size)

    @classmethod
    def get_batch_size(cls, batch: t.Any, batch_dim: int) -> int:
        if isinstance(batch, list):
            return len(t.cast(t.List[t.Any], batch))
        return 1

    @classmethod
    def from_payload(cls, payload: Payload) -> t.Any:
        return fixed_torch_loads(payload.data)


class DataContainerRegistry:
    CONTAINER_SINGLE_TYPE_MAP: t.Dict[
        LazyType[t.Any], t.Type[DataContainer[t.Any, t.Any]]
    ] = dict()
    CONTAINER_BATCH_TYPE_MAP: t.Dict[
        LazyType[t.Any], type[DataContainer[t.Any, t.Any]]
    ] = dict()

    @classmethod
    def register_container(
        cls,
        single_type: LazyType[t.Any] | type,
        batch_type: LazyType[t.Any] | type,
        container_cls: t.Type[DataContainer[t.Any, t.Any]],
    ):
        single_type = LazyType.from_type(single_type)
        batch_type = LazyType.from_type(batch_type)

        cls.CONTAINER_BATCH_TYPE_MAP[batch_type] = container_cls
        cls.CONTAINER_SINGLE_TYPE_MAP[single_type] = container_cls

    @classmethod
    def find_by_single_type(
        cls, type_: t.Type[SingleType] | LazyType[SingleType]
    ) -> t.Type[DataContainer[SingleType, t.Any]]:
        typeref = LazyType.from_type(type_)
        if typeref in cls.CONTAINER_SINGLE_TYPE_MAP:
            return cls.CONTAINER_SINGLE_TYPE_MAP[typeref]
        for klass in cls.CONTAINER_SINGLE_TYPE_MAP:
            if klass.issubclass(type_):
                return cls.CONTAINER_SINGLE_TYPE_MAP[klass]
        return DefaultContainer

    @classmethod
    def find_by_batch_type(
        cls, type_: t.Type[BatchType] | LazyType[BatchType]
    ) -> t.Type[DataContainer[t.Any, BatchType]]:
        typeref = LazyType.from_type(type_)
        if typeref in cls.CONTAINER_BATCH_TYPE_MAP:
            return cls.CONTAINER_BATCH_TYPE_MAP[typeref]
        for klass in cls.CONTAINER_BATCH_TYPE_MAP:
            if klass.issubclass(type_):
                return cls.CONTAINER_BATCH_TYPE_MAP[klass]
        return DefaultContainer

    @classmethod
    def find_by_name(cls, name: str) -> t.Type[DataContainer[t.Any, t.Any]]:
        for container_cls in cls.CONTAINER_BATCH_TYPE_MAP.values():
            if container_cls.__name__ == name:
                return container_cls
        if name == DefaultContainer.__name__:
            return DefaultContainer
        raise ValueError(f"can not find specified container class by name {name}")


def register_builtin_containers():
    DataContainerRegistry.register_container(
        LazyType("numpy", "ndarray"), LazyType("numpy", "ndarray"), NdarrayContainer
    )
    DataContainerRegistry.register_container(
        LazyType("pandas.core.series", "Series"),
        LazyType("pandas.core.frame", "DataFrame"),
        PandasDataFrameContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("pandas.core.frame", "DataFrame"),
        LazyType("pandas.core.frame", "DataFrame"),
        PandasDataFrameContainer,
    )

    DataContainerRegistry.register_container(
        LazyType("tritonclient.grpc", "InferInput"),
        LazyType("tritonclient.grpc", "InferInput"),
        TritonInferInputDataContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("tritonclient.grpc.aio", "InferInput"),
        LazyType("tritonclient.grpc.aio", "InferInput"),
        TritonInferInputDataContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("tritonclient.http", "InferInput"),
        LazyType("tritonclient.http", "InferInput"),
        TritonInferInputDataContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("tritonclient.http.aio", "InferInput"),
        LazyType("tritonclient.http.aio", "InferInput"),
        TritonInferInputDataContainer,
    )
    DataContainerRegistry.register_container(
        LazyType("PIL.Image", "Image"),
        LazyType("PIL.Image", "Image"),
        PILImageContainer,
    )
    DataContainerRegistry.register_container(Params, Params, ParamsContainer)
    DataContainerRegistry.register_container(Payload, Payload, PayloadContainer)


register_builtin_containers()


class AutoContainer(DataContainer[t.Any, t.Any]):
    @classmethod
    def to_payload(cls, batch: t.Any, batch_dim: int) -> Payload:
        container_cls: t.Type[DataContainer[t.Any, t.Any]] = (
            DataContainerRegistry.find_by_batch_type(type(batch))
        )
        return container_cls.to_payload(batch, batch_dim)

    @classmethod
    def get_batch_size(cls, batch: Any, batch_dim: int) -> int:
        container_cls: t.Type[DataContainer[t.Any, t.Any]] = (
            DataContainerRegistry.find_by_batch_type(type(batch))
        )
        return container_cls.get_batch_size(batch, batch_dim)

    @classmethod
    def from_payload(cls, payload: Payload) -> t.Any:
        container_cls = DataContainerRegistry.find_by_name(payload.container)
        return container_cls.from_payload(payload)

    @t.overload
    @classmethod
    def to_triton_payload(
        cls,
        inp: tritongrpcclient.InferInput,
        meta: tritongrpcclient.service_pb2.ModelMetadataResponse.TensorMetadata,
        _use_http_client: t.Literal[False] = ...,
    ) -> tritongrpcclient.InferInput: ...

    @t.overload
    @classmethod
    def to_triton_payload(
        cls,
        inp: tritonhttpclient.InferInput,
        meta: dict[str, t.Any],
        _use_http_client: t.Literal[True] = ...,
    ) -> tritongrpcclient.InferInput: ...

    @classmethod
    def to_triton_payload(
        cls, inp: t.Any, meta: t.Any, _use_http_client: bool = False
    ) -> t.Any:
        """
        Convert given input types to a Triton payload.

        Implementation of this method should always return a InferInput that has
        data of a Numpy array.
        """
        if _use_http_client:
            return DataContainerRegistry.find_by_single_type(
                type(inp)
            ).to_triton_http_payload(inp, meta)
        else:
            return DataContainerRegistry.find_by_single_type(
                type(inp)
            ).to_triton_grpc_payload(inp, meta)

    @classmethod
    def batches_to_batch(
        cls, batches: t.Sequence[BatchType], batch_dim: int = 0
    ) -> tuple[BatchType, list[int]]:
        container_cls: t.Type[DataContainer[t.Any, t.Any]] = (
            DataContainerRegistry.find_by_batch_type(type(batches[0]))
        )
        return container_cls.batches_to_batch(batches, batch_dim)

    @classmethod
    def batch_to_batches(
        cls, batch: BatchType, indices: t.Sequence[int], batch_dim: int = 0
    ) -> list[BatchType]:
        container_cls: t.Type[DataContainer[t.Any, t.Any]] = (
            DataContainerRegistry.find_by_batch_type(type(batch))
        )
        return container_cls.batch_to_batches(batch, indices, batch_dim)

    @classmethod
    def batch_to_payloads(
        cls,
        batch: t.Any,
        indices: t.Sequence[int],
        batch_dim: int = 0,
    ) -> list[Payload]:
        container_cls: t.Type[DataContainer[t.Any, t.Any]] = (
            DataContainerRegistry.find_by_batch_type(type(batch))
        )
        return container_cls.batch_to_payloads(batch, indices, batch_dim)

    @classmethod
    def from_batch_payloads(
        cls, payloads: t.Sequence[Payload], batch_dim: int = 0
    ) -> tuple[t.Any, list[int]]:
        container_cls = DataContainerRegistry.find_by_name(payloads[0].container)
        return container_cls.from_batch_payloads(payloads, batch_dim)



from __future__ import annotations

import logging
import typing as t
from types import ModuleType

import bentoml

from ...exceptions import MissingDependencyException
from ...exceptions import NotFound
from ..models.model import Model
from ..models.model import ModelContext
from ..models.model import ModelSignature
from ..models.model import PartialKwargsModelOptions as ModelOptions
from ..runner.utils import Params
from ..tag import Tag
from ..types import LazyType
from ..utils.lazy_loader import LazyLoader
from ..utils.pkg import get_pkg_version
from .common.pytorch import PyTorchTensorContainer  # noqa # type: ignore
from .common.pytorch import inference_mode_ctx
from .common.pytorch import torch

try:
    import detectron2.checkpoint as Checkpoint
    import detectron2.config as Config
    import detectron2.engine as Engine
    import detectron2.modeling as Modeling
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'detectron2' is required in order to use module 'bentoml.detectron'. Install detectron2 with 'pip install 'git+https://github.com/facebookresearch/detectron2.git''."
    )

if t.TYPE_CHECKING:
    import torch.nn as nn

    from .. import external_typing as ext
    from ..models.model import ModelSignaturesType
else:
    nn = LazyLoader("nn", globals(), "torch.nn")


__all__ = ["load_model", "save_model", "get_runnable", "get"]

MODULE_NAME = "bentoml.detectron"
API_VERSION = "v1"
MODEL_FILENAME = "saved_model"
DETECTOR_EXTENSION = ".pth"


logger = logging.getLogger(__name__)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def get(tag_like: str | Tag) -> Model:
    """
    Get the BentoML model with the given tag.

    Args:
        tag_like: The tag of the model to retrieve from the model store.

    Returns:
        :obj:`~bentoml.Model`: A BentoML :obj:`~bentoml.Model` with the matching tag.

    Example:

    .. code-block:: python

       import bentoml
       # target model must be from the BentoML model store
       model = bentoml.detectron2.get("en_reader:latest")
    """
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, not loading with {MODULE_NAME}."
        )
    return model


def load_model(
    bento_model: str | Tag | Model, device_id: str = "cpu"
) -> Engine.DefaultPredictor | nn.Module:
    """
    Load the detectron2 model from BentoML local model store with given name.

    Args:
        bento_model: Either the tag of the model to get from the store,
                     or a BentoML :class:`~bentoml.Model` instance to load the
                     model from.
        device_id: The device to load the model to. Default to "cpu".

    Returns:
        One of the following:
        - ``detectron2.engine.DefaultPredictor`` if the the checkpointables is saved as a Predictor.
        - ``torch.nn.Module`` if the checkpointables is saved as a nn.Module

    Example:

    .. code-block:: python

        import bentoml
        predictor = bentoml.detectron2.load_model('predictor:latest')

        model = bentoml.detectron2.load_model('model:latest')
    """
    if not isinstance(bento_model, Model):
        bento_model = get(bento_model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, not loading with {MODULE_NAME}."
        )

    cfg = bento_model.custom_objects["config"]
    cfg.MODEL.DEVICE = device_id

    metadata = bento_model.info.metadata
    if metadata.get("_is_predictor", True):
        return Engine.DefaultPredictor(cfg)
    else:
        model = Modeling.build_model(cfg)
        model.to(device).eval()
        Checkpoint.DetectionCheckpointer(model).load(
            bento_model.path_of(f"{MODEL_FILENAME}{DETECTOR_EXTENSION}")
        )
        return model


def save_model(
    name: Tag | str,
    checkpointables: Engine.DefaultPredictor | nn.Module,
    config: Config.CfgNode | None = None,
    *,
    signatures: ModelSignaturesType | None = None,
    labels: dict[str, str] | None = None,
    custom_objects: dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: dict[str, t.Any] | None = None,
) -> bentoml.Model:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name: Name for given model instance. This should pass Python identifier check.
        checkpointables: The model instance to be saved. Could be a ``detectron2.engine.DefaultPredictor`` or a ``torch.nn.Module``.
        config: Optional ``CfgNode`` for the model. Required when checkpointables is a ``torch.nn.Module``.
        signatures: Methods to expose for running inference on the target model. Signatures are used for creating :obj:`~bentoml.Runner` instances when serving model with :obj:`~bentoml.Service`
        labels: User-defined labels for managing models, e.g. ``team=nlp``, ``stage=dev``.
        custom_objects: Custom objects to be saved with the model. An example is ``{"my-normalizer": normalizer}``.
                        Custom objects are currently serialized with cloudpickle, but this implementation is subject to change.
        external_modules: user-defined additional python modules to be saved alongside the model or custom objects,
                          e.g. a tokenizer module, preprocessor module, model configuration module
        metadata: Custom metadata for given model.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format ``name:version`` where ``name`` is the user-defined model's name, and a generated ``version``.

    Examples:

    .. tab-set::

       .. tab-item:: ModelZoo and CfgNode

          .. code-block:: python

              import bentoml
              import detectron2
              from detectron2 import model_zoo
              from detectron2.config import get_cfg
              from detectron2.modeling import build_model

              model_url = "COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml"
              cfg = get_cfg()
              cfg.merge_from_file(model_zoo.get_config_file(model_url))
              # set threshold for this model
              cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5
              cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url(model_url)
              cloned = cfg.clone()
              cloned.MODEL.DEVICE = "cpu"

              bento_model = bentoml.detectron2.save_model('mask_rcnn', build_model(cloned), config=cloned)

       .. tab-item:: Predictor

          .. code-block:: python

              import bentoml
              import detectron2
              from detectron2.engine import DefaultPredictor
              from detectron2 import model_zoo
              from detectron2.config import get_cfg
              from detectron2.modeling import build_model

              model_url = "COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml"
              cfg = get_cfg()
              cfg.merge_from_file(model_zoo.get_config_file(model_url))
              # set threshold for this model
              cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5
              cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url(model_url)
              cloned = cfg.clone()
              cloned.MODEL.DEVICE = "cpu"

              predictor = DefaultPredictor(cloned)
              bento_model = bentoml.detectron2.save_model('mask_rcnn', predictor)
    """  # noqa
    context = ModelContext(
        framework_name="detectron2",
        framework_versions={"detectron2": get_pkg_version("detectron2")},
    )

    if signatures is None:
        signatures = {"__call__": {"batchable": False}}
        logger.info(
            'Using the default model signature for Detectron (%s) for model "%s".',
            signatures,
            name,
        )
    if metadata is None:
        metadata = {}

    metadata["_is_predictor"] = isinstance(checkpointables, Engine.DefaultPredictor)

    if custom_objects is None:
        custom_objects = {}

    if isinstance(checkpointables, nn.Module):
        if config is None:
            raise ValueError(
                "config is required when 'checkpointables' is a derived 'torch.nn.Module'."
            )
        model = checkpointables
        model.eval()
    else:
        model = checkpointables.model
        if config is not None:
            logger.warning(
                "config is ignored when 'checkpointables' is a 'DefaultPredictor'."
            )
        config = t.cast(Config.CfgNode, checkpointables.cfg)

    custom_objects["config"] = config

    with bentoml.models._create(  # type: ignore
        name,
        module=MODULE_NAME,
        api_version=API_VERSION,
        labels=labels,
        context=context,
        options=ModelOptions(),
        signatures=signatures,
        custom_objects=custom_objects,
        external_modules=external_modules,
        metadata=metadata,
    ) as bento_model:
        checkpointer = Checkpoint.Checkpointer(model, save_dir=bento_model.path)
        checkpointer.save(MODEL_FILENAME)
        return bento_model


def get_runnable(bento_model: bentoml.Model) -> type[bentoml.Runnable]:
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    is_predictor = bento_model.info.metadata.get("_is_predictor", True)
    partial_kwargs = t.cast(ModelOptions, bento_model.info.options).partial_kwargs

    class Detectron2Runnable(bentoml.Runnable):
        SUPPORTED_RESOURCES = ("nvidia.com/gpu", "cpu")
        SUPPORTS_CPU_MULTI_THREADING = True

        def __init__(self):
            super().__init__()
            if torch.cuda.is_available():
                self.device_id = "cuda"
                torch.set_default_tensor_type("torch.cuda.FloatTensor")
            else:
                self.device_id = "cpu"

            self.model = load_model(bento_model, device_id=self.device_id)

            if not is_predictor:
                assert isinstance(self.model, torch.nn.Module)
                # This predictor is a torch.nn.Module
                self.model.train(False)

            self.is_predictor = is_predictor

            self.predict_fns: dict[str, t.Callable[..., t.Any]] = {}
            for method_name in bento_model.info.signatures:
                self.predict_fns[method_name] = getattr(self.model, method_name)

    def add_runnable_method(method_name: str, options: ModelSignature):
        def _run(self: Detectron2Runnable, *args: t.Any, **kwargs: t.Any) -> t.Any:
            method_partial_kwargs = partial_kwargs.get(method_name, {})

            params = Params(*args, **kwargs)

            if self.is_predictor:
                return self.predict_fns[method_name](
                    *params.args, **dict(method_partial_kwargs, **params.kwargs)
                )

            def mapping(item: ext.NpNDArray | torch.Tensor) -> t.Any:
                if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(item):
                    return torch.Tensor(item, device=self.device_id)
                elif isinstance(item, torch.Tensor):
                    return item.to(self.device_id)
                else:
                    return item

            with inference_mode_ctx():
                params = params.map(mapping)
                return self.predict_fns[method_name](
                    [{"image": image} for image in params.args],
                    **dict(method_partial_kwargs, **params.kwargs),
                )

        Detectron2Runnable.add_method(
            _run,
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )

    for method_name, options in bento_model.info.signatures.items():
        add_runnable_method(method_name, options)

    return Detectron2Runnable



from __future__ import annotations

import abc
import datetime
import enum
import textwrap
import typing as t
import uuid

import attrs
from starlette.requests import Request
from starlette.responses import Response

from .serde import JSONSerde
from .serde import Serde

Ti = t.TypeVar("Ti")
To = t.TypeVar("To")

if t.TYPE_CHECKING:
    from aiosqlite import Connection


class ResultStatus(enum.Enum):
    IN_PROGRESS = "in_progress"
    SUCCESS = "success"
    FAILURE = "failure"
    CANCELLED = "cancelled"


@attrs.frozen
class ResultRow:
    task_id: str
    name: str
    status: ResultStatus
    created_at: datetime.datetime
    executed_at: datetime.datetime | None

    def to_json(self) -> dict[str, t.Any]:
        return {
            "task_id": self.task_id,
            "status": self.status.value,
            "created_at": self.created_at.isoformat(),
            "executed_at": self.executed_at.isoformat() if self.executed_at else None,
        }


@attrs.frozen
class CompletedResultRow(ResultRow, t.Generic[Ti, To]):
    input: Ti
    result: To
    completed_at: datetime.datetime | None

    def to_json(self) -> dict[str, t.Any]:
        return {
            **super().to_json(),
            "completed_at": self.completed_at.isoformat()
            if self.completed_at
            else None,
        }


class ResultStore(abc.ABC, t.Generic[Ti, To]):
    @abc.abstractmethod
    async def get_status(self, task_id: str) -> ResultRow:
        raise NotImplementedError

    @abc.abstractmethod
    async def get(self, task_id: str) -> CompletedResultRow[Ti, To]:
        raise NotImplementedError

    async def get_or_none(self, task_id: str) -> t.Optional[ResultRow[Ti, To]]:
        try:
            return await self.get(task_id)
        except (KeyError, RuntimeError):
            return None

    @abc.abstractmethod
    async def new_entry(self, name: str, input: Ti) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    async def set_result(self, task_id: str, result: To, status: ResultStatus) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def set_status(self, task_id: str, status: ResultStatus) -> None:
        raise NotImplementedError


class Sqlite3Store(ResultStore[Request, Response]):
    RESULT_RETENTION = 60 * 60 * 24  # retain results for 24 hours

    def __init__(self, db_file: str, serializer: Serde | None = None) -> None:
        self._conn = self._connect(db_file)
        if serializer is None:
            serializer = JSONSerde()
        self.serializer = serializer

    def _connect(self, db_file: str) -> Connection:
        import sqlite3

        import aiosqlite

        return aiosqlite.connect(
            db_file, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES
        )

    async def __aenter__(self) -> "t.Self":
        self._conn = await self._conn
        await self._init_db()
        return self

    async def __aexit__(self, *_: t.Any) -> None:
        await self._conn.close()

    async def _init_db(self) -> None:
        await self._conn.execute(
            textwrap.dedent("""\
            CREATE TABLE IF NOT EXISTS result (
                task_id TEXT PRIMARY KEY,
                name TEXT,
                input BLOB,
                status TEXT,
                result BLOB,
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                executed_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                completed_at TIMESTAMP DEFAULT NULL
            )
            """)
        )
        await self._conn.commit()

    async def new_entry(self, name: str, input: Request) -> str:
        task_id = uuid.uuid4().hex
        await self._conn.execute(
            "INSERT INTO result (task_id, name, input, status) VALUES (?, ?, ?, ?)",
            (
                task_id,
                name,
                await self.serializer.serialize_request(input),
                ResultStatus.IN_PROGRESS.value,
            ),
        )
        await self._conn.commit()
        return task_id

    async def get(self, task_id: str) -> CompletedResultRow[Request, Response]:
        result = await self._conn.execute(
            "SELECT name, input, status, result, created_at, executed_at, completed_at "
            "FROM result WHERE task_id = ?",
            (task_id,),
        )
        row = await result.fetchone()
        if row is None:
            raise KeyError(task_id)
        if row[6] is None:
            raise RuntimeError(f"Task {task_id} is not completed")
        return CompletedResultRow(
            task_id,
            row[0],
            ResultStatus(row[2]),
            row[4],
            row[5],
            await self.serializer.deserialize_request(row[1]),
            await self.serializer.deserialize_response(row[3]),
            row[6],
        )

    async def get_status(self, task_id: str) -> ResultRow:
        result = await self._conn.execute(
            "SELECT name, status, created_at, executed_at "
            "FROM result WHERE task_id = ?",
            (task_id,),
        )
        row = await result.fetchone()
        if row is None:
            raise KeyError(task_id)

        return ResultRow(
            task_id,
            row[0],
            ResultStatus(row[1]),
            row[2],
            row[3],
        )

    async def set_status(self, task_id: str, status: ResultStatus) -> None:
        await self._conn.execute(
            "UPDATE result SET status = ?, updated_at = ? WHERE task_id = ? AND status = ?",
            (
                status.value,
                datetime.datetime.now(tz=datetime.timezone.utc),
                task_id,
                ResultStatus.IN_PROGRESS.value,
            ),
        )
        await self._conn.commit()

    async def set_result(
        self, task_id: str, result: Response, status: ResultStatus
    ) -> None:
        await self._conn.execute(
            "UPDATE result SET status = ?, result = ?, completed_at = ? WHERE task_id = ?",
            (
                status.value,
                await self.serializer.serialize_response(result),
                datetime.datetime.now(tz=datetime.timezone.utc),
                task_id,
            ),
        )
        # delete older results
        await self._conn.execute(
            "DELETE FROM result WHERE completed_at IS NOT NULL AND completed_at < ?",
            (
                datetime.datetime.now(tz=datetime.timezone.utc)
                - datetime.timedelta(seconds=self.RESULT_RETENTION),
            ),
        )
        await self._conn.commit()



from __future__ import annotations

from ._internal.frameworks.diffusers_runners import stable_diffusion
from ._internal.frameworks.diffusers_runners import stable_diffusion_xl

__all__ = ["stable_diffusion", "stable_diffusion_xl"]



from __future__ import annotations

import io
import typing as t

import attr

from bentoml.exceptions import BentoMLException

if t.TYPE_CHECKING:

    class SSEArgs(t.TypedDict, total=False):
        data: str
        event: str
        id: str
        retry: int


@attr.s(auto_attribs=True)
class SSE:
    data: str
    event: str | None = None
    id: str | None = None
    retry: int | None = None

    def marshal(self) -> str:
        with io.StringIO() as buffer:
            if self.event:
                if "\n" in self.event:
                    raise BentoMLException("SSE event name should not contain new line")
                buffer.write(f"event: {self.event}\n")
            if self.id is not None:
                buffer.write(f"id: {self.id}\n")
            if self.retry is not None:
                buffer.write(f"retry: {self.retry}\n")
            for line in self.data.split("\n"):
                buffer.write(f"data: {line}\n")
            buffer.write("\n")
            return buffer.getvalue()

    @classmethod
    def _read_sse(cls, buffer: io.StringIO) -> SSE:
        event: SSEArgs = {}
        data_buffer: io.StringIO | None = None
        first_data_line = True

        while True:
            line = buffer.readline()
            if not line:
                break
            if line == "\n":
                break
            if line.startswith("data: "):
                if first_data_line:
                    first_data_line = False
                    event["data"] = line[6:-1]
                else:
                    if data_buffer is None:
                        # only init data_buffer when there is more than one data line
                        data_buffer = io.StringIO()
                        data_buffer.write(event.get("data", ""))
                    data_buffer.write("\n")
                    data_buffer.write(line[6:-1])
            if line.startswith("event: "):
                event["event"] = line[7:].strip()
            if line.startswith("id: "):
                event["id"] = line[4:].strip()
            if line.startswith("retry: "):
                event["retry"] = int(line[7:].strip())

        if data_buffer is not None:
            event["data"] = data_buffer.getvalue()
            data_buffer.close()

        return SSE(**event)

    @classmethod
    async def from_iterator(
        cls,
        async_iterator: t.AsyncIterator[str],
    ) -> t.AsyncGenerator[SSE, None]:
        with io.StringIO() as buffer:
            read_cursor = 0
            last_chunk = ""
            async for chunk in async_iterator:
                buffer.write(chunk)
                if (
                    "\n\n" in chunk
                    or chunk
                    and chunk[0] == "\n"
                    and last_chunk
                    and last_chunk[-1] == "\n"
                ):
                    buffer.seek(read_cursor)
                    yield cls._read_sse(buffer)
                    read_cursor = buffer.tell()
                last_chunk = chunk



from __future__ import annotations

import dataclasses
import json
import logging
import typing as t

import attr
from starlette.requests import Request
from starlette.responses import Response

from ...exceptions import BadInput
from ...exceptions import InvalidArgument
from ..service.openapi import REF_PREFIX
from ..service.openapi import SUCCESS_DESCRIPTION
from ..service.openapi.specification import MediaType
from ..service.openapi.specification import Schema
from ..types import LazyType
from ..utils.cattr import bentoml_cattr
from ..utils.http import set_cookies
from ..utils.lazy_loader import LazyLoader
from ..utils.pkg import pkg_version_info
from .base import IODescriptor

EXC_MSG = "'pydantic' must be installed to use 'pydantic_model'. Install with 'pip install bentoml[io-json]'."

if t.TYPE_CHECKING:
    from types import UnionType

    import pydantic
    import pydantic.schema as schema

    if pkg_version_info("pydantic")[0] >= 2:
        import pydantic.json_schema as jschema

    from google.protobuf import message as _message
    from google.protobuf import struct_pb2
    from typing_extensions import Self

    from ..context import ServiceContext as Context
    from .base import OpenAPIResponse

else:
    pydantic = LazyLoader("pydantic", globals(), "pydantic", exc_msg=EXC_MSG)
    schema = LazyLoader("schema", globals(), "pydantic.schema", exc_msg=EXC_MSG)
    jschema = LazyLoader(
        "jschema", globals(), "pydantic.json_schema", exc_msg="Pydantic v2 is required."
    )
    # lazy load our proto generated.
    struct_pb2 = LazyLoader("struct_pb2", globals(), "google.protobuf.struct_pb2")
    # lazy load numpy for processing ndarray.
    np = LazyLoader("np", globals(), "numpy")


JSONType = t.Union[str, t.Dict[str, t.Any], "pydantic.BaseModel", None]

logger = logging.getLogger(__name__)


class DefaultJsonEncoder(json.JSONEncoder):
    def default(self, o: type) -> t.Any:
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(o):
            return o.tolist()
        if LazyType["ext.NpGeneric"]("numpy.generic").isinstance(o):
            return o.item()
        if LazyType["ext.PdDataFrame"]("pandas.DataFrame").isinstance(o):
            return o.to_dict()  # type: ignore
        if LazyType["ext.PdSeries"]("pandas.Series").isinstance(o):
            return o.to_dict()  # type: ignore
        if LazyType["pydantic.BaseModel"]("pydantic.BaseModel").isinstance(o):
            obj_dict = o.dict()
            if "__root__" in obj_dict:
                obj_dict = obj_dict.get("__root__")
            return obj_dict
        if attr.has(o):
            return bentoml_cattr.unstructure(o)
        return super().default(o)


class JSON(
    IODescriptor[JSONType], descriptor_id="bentoml.io.JSON", proto_fields=("json",)
):
    """
    :obj:`JSON` defines API specification for the inputs/outputs of a Service, where either
    inputs will be converted to or outputs will be converted from a JSON representation
    as specified in your API function signature.

    A sample service implementation:

    .. code-block:: python
       :caption: `service.py`

       from __future__ import annotations

       import typing
       from typing import TYPE_CHECKING
       from typing import Any
       from typing import Optional

       import bentoml
       from bentoml.io import NumpyNdarray
       from bentoml.io import JSON

       import numpy as np
       import pandas as pd
       from pydantic import BaseModel

       iris_clf_runner = bentoml.sklearn.get("iris_clf_with_feature_names:latest").to_runner()

       svc = bentoml.Service("iris_classifier_pydantic", runners=[iris_clf_runner])

       class IrisFeatures(BaseModel):
           sepal_len: float
           sepal_width: float
           petal_len: float
           petal_width: float

           # Optional field
           request_id: Optional[int]

           # Use custom Pydantic config for additional validation options
           class Config:
               extra = 'forbid'


       input_spec = JSON(pydantic_model=IrisFeatures)

       @svc.api(input=input_spec, output=NumpyNdarray())
       def classify(input_data: IrisFeatures) -> NDArray[Any]:
           if input_data.request_id is not None:
               print("Received request ID: ", input_data.request_id)

           input_df = pd.DataFrame([input_data.dict(exclude={"request_id"})])
           return iris_clf_runner.run(input_df)

    Users then can then serve this service with :code:`bentoml serve`:

    .. code-block:: bash

        % bentoml serve ./service.py:svc --reload

    Users can then send requests to the newly started services with any client:

    .. tab-set::

        .. tab-item:: Bash

            .. code-block:: bash

               % curl -X POST -H "content-type: application/json" \\
                   --data '{"sepal_len": 6.2, "sepal_width": 3.2, "petal_len": 5.2, "petal_width": 2.2}' \\
                   http://127.0.0.1:3000/classify

               # [2]%

        .. tab-item:: Python

            .. code-block:: python
               :caption: `request.py`

                import requests

                requests.post(
                    "http://0.0.0.0:3000/predict",
                    headers={"content-type": "application/json"},
                    data='{"sepal_len": 6.2, "sepal_width": 3.2, "petal_len": 5.2, "petal_width": 2.2}'
                ).text

    Args:
        pydantic_model: Pydantic model schema. When used, inference API callback
                        will receive an instance of the specified ``pydantic_model`` class.
        json_encoder: JSON encoder class. By default BentoML implements a custom JSON encoder that
                      provides additional serialization supports for numpy arrays, pandas dataframes,
                      dataclass-like (`attrs <https://www.attrs.org/en/stable/>`_, dataclass, etc.).
                      If you wish to use a custom encoder, make sure to support the aforementioned object.

    Returns:
        :obj:`JSON`: IO Descriptor that represents JSON format.
    """

    # default mime type is application/json
    _mime_type = "application/json"

    def __init__(
        self,
        *,
        pydantic_model: type[pydantic.BaseModel] | None = None,
        validate_json: bool | None = None,
        json_encoder: type[json.JSONEncoder] = DefaultJsonEncoder,
    ):
        if pydantic_model is not None:
            assert issubclass(pydantic_model, pydantic.BaseModel), (
                "'pydantic_model' must be a subclass of 'pydantic.BaseModel'."
            )

        self._pydantic_model = pydantic_model
        self._json_encoder = json_encoder

        # Remove validate_json in version 1.0.2
        if validate_json is not None:
            logger.warning(
                "'validate_json' option from 'bentoml.io.JSON' has been deprecated. Use a Pydantic model to specify validation options instead."
            )

    def _from_sample(self, sample: JSONType) -> JSONType:
        """
        Create a :class:`~bentoml._internal.io_descriptors.json.JSON` IO Descriptor from given inputs.

        Args:
            sample: A JSON-like datatype, which can be either dict, str, list.
                    ``sample`` will also accepting a Pydantic model.

                    .. code-block:: python

                        from pydantic import BaseModel

                        class IrisFeatures(BaseModel):
                            sepal_len: float
                            sepal_width: float
                            petal_len: float
                            petal_width: float

                        input_spec = JSON.from_sample(
                            IrisFeatures(sepal_len=1.0, sepal_width=2.0, petal_len=3.0, petal_width=4.0)
                        )

                        @svc.api(input=input_spec, output=NumpyNdarray())
                        async def predict(input: NDArray[np.int16]) -> NDArray[Any]:
                            return await runner.async_run(input)
            json_encoder: Optional JSON encoder.

        Returns:
            :class:`~bentoml._internal.io_descriptors.json.JSON`: IODescriptor from given users inputs.

        Example:

        .. code-block:: python
           :caption: `service.py`

           from __future__ import annotations

           import bentoml
           from typing import Any
           from bentoml.io import JSON

           input_spec = JSON.from_sample({"Hello": "World", "foo": "bar"})
           @svc.api(input=input_spec, output=JSON())
           async def predict(input: dict[str, Any]) -> dict[str, Any]:
               return await runner.async_run(input)

        Raises:
            :class:`BadInput`: Given sample is not a valid JSON string, bytes, or supported nest types.
        """
        if LazyType["pydantic.BaseModel"]("pydantic.BaseModel").isinstance(sample):
            self._pydantic_model = sample.__class__
        elif isinstance(sample, str):
            try:
                sample = json.loads(sample)
            except json.JSONDecodeError as e:
                raise BadInput(
                    f"Unable to parse JSON string. Please make sure the input is a valid JSON string: {e}"
                ) from None
        elif isinstance(sample, bytes):
            try:
                sample = json.loads(sample.decode())
            except json.JSONDecodeError as e:
                raise BadInput(
                    f"Unable to parse JSON bytes. Please make sure the input is a valid JSON bytes: {e}"
                ) from None
        elif not isinstance(sample, (dict, list)):
            raise BadInput(
                f"Unable to infer JSON type from sample: {sample}. Please make sure the input is a valid JSON object."
            )
        return sample

    def to_spec(self) -> dict[str, t.Any]:
        return {
            "id": self.descriptor_id,
            "args": {
                "has_pydantic_model": self._pydantic_model is not None,
                "has_json_encoder": self._json_encoder is not DefaultJsonEncoder,
            },
        }

    @classmethod
    def from_spec(cls, spec: dict[str, t.Any]) -> Self:
        if "args" not in spec:
            raise InvalidArgument(f"Missing args key in JSON spec: {spec}")
        if "has_pydantic_model" in spec["args"] and spec["args"]["has_pydantic_model"]:
            logger.warning(
                "BentoML does not support loading pydantic models from URLs; output will be a normal dictionary."
            )
        if "has_json_encoder" in spec["args"] and spec["args"]["has_json_encoder"]:
            logger.warning(
                "BentoML does not support loading JSON encoders from URLs; output will be a normal dictionary."
            )

        return cls()

    def input_type(self) -> UnionType:
        return JSONType

    def openapi_schema(self) -> Schema:
        if not self._pydantic_model:
            return Schema(type="object")

        # returns schemas from pydantic_model.
        if pkg_version_info("pydantic")[0] >= 2:
            json_schema = jschema.model_json_schema(
                self._pydantic_model, ref_template=REF_PREFIX + "{model}"
            )
            # NOTE: we don't need def here, as these will be available in openapi.components.
            if "$defs" in json_schema:
                json_schema.pop("$defs", None)
            return Schema(**json_schema)
        else:
            return Schema(
                **schema.model_process_schema(
                    self._pydantic_model,
                    model_name_map=schema.get_model_name_map(
                        schema.get_flat_models_from_model(self._pydantic_model)
                    ),
                    ref_prefix=REF_PREFIX,
                )[0]
            )

    def openapi_components(self) -> dict[str, t.Any] | None:
        if not self._pydantic_model:
            return {}

        from ..service.openapi.utils import pydantic_components_schema

        return {"schemas": pydantic_components_schema(self._pydantic_model)}

    def openapi_example(self):
        if self.sample is not None:
            if LazyType["pydantic.BaseModel"]("pydantic.BaseModel").isinstance(
                self.sample
            ):
                if pkg_version_info("pydantic")[0] >= 2:
                    return self.sample.model_dump()
                else:
                    return self.sample.dict()
            elif isinstance(self.sample, (str, list)):
                return json.dumps(
                    self.sample,
                    cls=self._json_encoder,
                    ensure_ascii=False,
                    allow_nan=False,
                    indent=None,
                    separators=(",", ":"),
                )
            elif isinstance(self.sample, dict):
                return self.sample

    def openapi_request_body(self) -> dict[str, t.Any]:
        return {
            "content": {
                self._mime_type: MediaType(
                    schema=self.openapi_schema(), example=self.openapi_example()
                )
            },
            "required": True,
            "x-bentoml-io-descriptor": self.to_spec(),
        }

    def openapi_responses(self) -> OpenAPIResponse:
        return {
            "description": SUCCESS_DESCRIPTION,
            "content": {
                self._mime_type: MediaType(
                    schema=self.openapi_schema(), example=self.openapi_example()
                )
            },
            "x-bentoml-io-descriptor": self.to_spec(),
        }

    async def from_http_request(self, request: Request) -> JSONType:
        json_str = await request.body()
        try:
            json_obj = json.loads(json_str)
        except json.JSONDecodeError as e:
            raise BadInput(f"Invalid JSON input received: {e}") from None

        if self._pydantic_model:
            try:
                if pkg_version_info("pydantic")[0] >= 2:
                    pydantic_model = self._pydantic_model.model_validate(json_obj)
                else:
                    pydantic_model = self._pydantic_model.parse_obj(json_obj)
                return pydantic_model
            except pydantic.ValidationError as e:
                raise BadInput(f"Invalid JSON input received: {e}") from None
        else:
            return json_obj

    async def to_http_response(
        self, obj: JSONType | pydantic.BaseModel, ctx: Context | None = None
    ):
        # This is to prevent cases where custom JSON encoder is used.
        if LazyType["pydantic.BaseModel"]("pydantic.BaseModel").isinstance(obj):
            if pkg_version_info("pydantic")[0] >= 2:
                obj = obj.model_dump()
            else:
                obj = obj.dict()

        json_str = (
            json.dumps(
                obj,
                cls=self._json_encoder,
                ensure_ascii=False,
                allow_nan=False,
                indent=None,
                separators=(",", ":"),
            )
            if obj is not None
            else None
        )

        if ctx is not None:
            res = Response(
                json_str,
                media_type=self._mime_type,
                headers=ctx.response.metadata,  # type: ignore (bad starlette types)
                status_code=ctx.response.status_code,
            )
            set_cookies(res, ctx.response.cookies)
            return res
        else:
            return Response(json_str, media_type=self._mime_type)

    async def from_proto(self, field: struct_pb2.Value | bytes) -> JSONType:
        from google.protobuf.json_format import MessageToDict

        if isinstance(field, bytes):
            content = field
            if self._pydantic_model:
                try:
                    if pkg_version_info("pydantic")[0] >= 2:
                        return self._pydantic_model.model_validate_json(content)
                    else:
                        return self._pydantic_model.parse_raw(content)
                except pydantic.ValidationError as e:
                    raise BadInput(f"Invalid JSON input received: {e}") from None
            try:
                parsed = json.loads(content)
            except json.JSONDecodeError as e:
                raise BadInput(f"Invalid JSON input received: {e}") from None
        else:
            assert isinstance(field, struct_pb2.Value)
            parsed = MessageToDict(field, preserving_proto_field_name=True)

            if self._pydantic_model:
                try:
                    if pkg_version_info("pydantic")[0] >= 2:
                        return self._pydantic_model.model_validate(parsed)
                    else:
                        return self._pydantic_model.parse_obj(parsed)
                except pydantic.ValidationError as e:
                    raise BadInput(f"Invalid JSON input received: {e}") from None
        return parsed

    async def to_proto(self, obj: JSONType) -> struct_pb2.Value:
        if LazyType["pydantic.BaseModel"]("pydantic.BaseModel").isinstance(obj):
            if pkg_version_info("pydantic")[0] >= 2:
                obj = obj.model_dump()
            else:
                obj = obj.dict()
        msg = struct_pb2.Value()
        return parse_dict_to_proto(obj, msg, json_encoder=self._json_encoder)


def parse_dict_to_proto(
    obj: JSONType,
    msg: _message.Message,
    json_encoder: type[json.JSONEncoder] = DefaultJsonEncoder,
) -> t.Any:
    if obj is None:
        # this function is an identity op for the msg if obj is None.
        return msg

    from google.protobuf.json_format import ParseDict

    if isinstance(obj, (dict, str, list, float, int, bool)):
        # ParseDict handles google.protobuf.Struct type
        # directly if given object has a supported type
        ParseDict(obj, msg)
    else:
        # If given object doesn't have a supported type, we will
        # use given JSON encoder to convert it to dictionary
        # and then parse it to google.protobuf.Struct.
        # Note that if a custom JSON encoder is used, it mustn't
        # take any arguments.
        ParseDict(json_encoder().default(obj), msg)
    return msg



from __future__ import annotations

import typing as t
from functools import lru_cache
from typing import TYPE_CHECKING

from bentoml.exceptions import InternalServerError
from bentoml.exceptions import InvalidArgument
from bentoml.exceptions import NotFound

from ...utils.lazy_loader import LazyLoader
from ...utils.pkg import pkg_version_info
from .specification import Schema

if TYPE_CHECKING:
    import pydantic
    import pydantic.schema as schema

    if pkg_version_info("pydantic")[0] >= 2:
        import pydantic.json_schema as jschema

    from bentoml.exceptions import BentoMLException

else:
    _exc_msg = (
        "Missing required dependency: 'pydantic'. Install with 'pip install pydantic'."
    )
    pydantic = LazyLoader("pydantic", globals(), "pydantic", exc_msg=_exc_msg)
    schema = LazyLoader("schema", globals(), "pydantic.schema", exc_msg=_exc_msg)
    jschema = LazyLoader(
        "jschema", globals(), "pydantic.json_schema", exc_msg="Pydantic v2 is required."
    )

REF_PREFIX = "#/components/schemas/"


def pydantic_components_schema(pydantic_model: type[pydantic.BaseModel]):
    if pkg_version_info("pydantic")[0] >= 2:
        return pydantic_v2_components_schema(pydantic_model)
    else:
        return pydantic_v1_components_schema(pydantic_model)


def pydantic_v2_components_schema(pydantic_model: type[pydantic.BaseModel]):
    json_schema = jschema.model_json_schema(
        pydantic_model, ref_template=REF_PREFIX + "{model}"
    )
    defs = json_schema.pop("$defs", None)
    components: dict[str, Schema] = {pydantic_model.__name__: Schema(**json_schema)}
    if defs is not None:
        # NOTE: This is a nested models, hence we will update the definitions
        components.update({k: Schema(**v) for k, v in defs.items()})
    return components


def pydantic_v1_components_schema(pydantic_model: type[pydantic.BaseModel]):
    flat_models = schema.get_flat_models_from_model(pydantic_model)
    model_name_map = schema.get_model_name_map(flat_models)

    # gets model definitions
    definitions: dict[str, dict[str, t.Any]] = {}
    for model in flat_models:
        m_schema, m_definitions, _ = schema.model_process_schema(
            model, model_name_map=model_name_map, ref_prefix=REF_PREFIX
        )
        definitions.update(m_definitions)
        model_name = model_name_map[model]
        definitions[model_name] = m_schema
    return {k: Schema(**definitions[k]) for k in sorted(definitions)}


@lru_cache(maxsize=1)
def exception_components_schema() -> dict[str, Schema]:
    return {
        schema.title: schema
        for ex in [InvalidArgument, NotFound, InternalServerError]
        for schema in exception_schema(ex)
    }


def exception_schema(ex: t.Type[BentoMLException]) -> t.Iterable[FilledExceptionSchema]:
    # convert BentoML exception to OpenAPI components schema
    error_properties = {
        "error": Schema(title="Message", type="string"),
        "detail": Schema(title="Error Detail", type="string"),
    }

    yield FilledExceptionSchema(
        title=ex.__name__,
        type="object",
        description=ex.error_code.phrase,
        properties=error_properties,
        required=["error"],
    )


class FilledExceptionSchema(Schema):
    title: str
    description: str



from __future__ import annotations

import contextlib
import logging
import typing as t

if t.TYPE_CHECKING:
    from _bentoml_sdk import Service
    from bentoml import Bento
    from bentoml import Tag


def load(
    bento_identifier: str | Tag | Bento,
    working_dir: str | None = None,
    reload: bool = False,
) -> Service[t.Any]:
    """An alias for `bentoml.load` that ensures to return a new-style service object."""
    from bentoml._internal.service.loader import load as load_service

    return t.cast(
        "Service[t.Any]",
        load_service(bento_identifier, working_dir=working_dir, reload=reload),
    )


def import_service(
    svc_import_path: str, working_dir: str | None = None, reload: bool = False
) -> Service[t.Any]:
    """An alias for `bentoml._internal.services.loader.import_service` that ensures
    to return a new-style service object.
    """
    from bentoml._internal.service.loader import import_service as import_service_impl

    return t.cast(
        "Service[t.Any]",
        import_service_impl(svc_import_path, working_dir=working_dir, reload=reload),
    )


@contextlib.contextmanager
def importing() -> t.Iterator[None]:
    """A context manager that suppresses the ImportError when importing runtime-specific modules.

    Example:

    .. code-block:: python

        with bentoml.importing():
            import torch
            import tensorflow
    """
    from bentoml._internal.context import server_context

    logger = logging.getLogger("bentoml.service")

    try:
        yield
    except ImportError as e:
        if server_context.worker_index is not None:
            raise
        logger.info(
            "Skipping import of module %s because it is not available in the current environment",
            e.name,
        )



import os
import pathlib
from urllib.parse import quote
from urllib.parse import unquote
from urllib.parse import urlparse
from urllib.request import url2pathname

import psutil


def path_to_uri(path: str) -> str:
    """
    Convert a path to a URI.

    Args:
        path: Path to convert to URI.

    Returns:
        URI string. (quoted, absolute)
    """
    path = os.path.abspath(path)
    if psutil.WINDOWS:
        return pathlib.PureWindowsPath(path).as_uri()
    if psutil.POSIX:
        return pathlib.PurePosixPath(path).as_uri()
    raise ValueError("Unsupported OS")


def uri_to_path(uri: str) -> str:
    """
    Convert a file URI to a path.

    Args:
        uri: URI to convert to path.

    Returns:
        Path string. (unquoted)
    """
    parsed = urlparse(uri)
    if parsed.scheme not in ("file", "filesystem", "unix"):
        raise ValueError("Unsupported URI scheme")
    host = "{0}{0}{mnt}{0}".format(os.path.sep, mnt=parsed.netloc)
    return os.path.normpath(os.path.join(host, url2pathname(unquote(parsed.path))))


def encode_path_for_uri(path: str) -> str:
    """Percent-encode non-URL characters in a path."""
    return quote(path.replace(os.sep, "/"))


def is_http_url(url: str) -> bool:
    return urlparse(url).scheme in {"http", "https"}



from __future__ import annotations

import collections.abc
import sys
import types
import typing as t

from typing_extensions import Literal

if t.TYPE_CHECKING:
    from PIL import Image

LITERAL_TYPES: set[type] = {Literal}
if hasattr(t, "Literal"):
    LITERAL_TYPES.add(t.Literal)
LIST_TYPES: set[type] = {list, t.List, t.Sequence, t.MutableSequence}
TUPLE_TYPES: set[type] = {tuple, t.Tuple}
SYNC_ITERATOR_TYPES: set[type] = {
    t.Iterator,
    t.Generator,
    collections.abc.Iterator,
    collections.abc.Generator,
}

ASYNC_ITERATOR_TYPES: set[type] = {
    t.AsyncIterator,
    t.AsyncGenerator,
    collections.abc.AsyncIterator,
    collections.abc.AsyncGenerator,
}


def get_origin(type_: t.Any) -> type:
    if hasattr(type_, "__origin__"):
        return type_.__origin__
    return t.get_origin(type_) or type_


def get_args(type_: t.Any) -> t.Tuple[t.Any, ...]:
    if hasattr(type_, "__args__"):
        return type_.__args__
    return t.get_args(type_)


def is_literal_type(typ_: t.Any) -> bool:
    return get_origin(typ_) in LITERAL_TYPES


def is_list_type(typ_: t.Any) -> bool:
    origin = get_origin(typ_)
    return issubclass(origin, list) or origin in LIST_TYPES


def is_tuple_type(typ_: t.Any) -> bool:
    return get_origin(typ_) in TUPLE_TYPES


def is_union_type(typ_: t.Any) -> bool:
    check_types = (t.Union,)
    if sys.version_info >= (3, 10):
        check_types += (types.UnionType,)
    return get_origin(typ_) in check_types


def is_iterator_type(typ_: t.Any) -> bool:
    return get_origin(typ_) in (SYNC_ITERATOR_TYPES | ASYNC_ITERATOR_TYPES)


def is_file_like(obj: t.Any) -> t.TypeGuard[t.BinaryIO]:
    return hasattr(obj, "read") and hasattr(obj, "seek")


def is_image_type(type_: type) -> t.TypeGuard[type[Image.Image]]:
    from bentoml._internal.io_descriptors.image import PIL

    return type_.__module__.startswith("PIL.") and issubclass(type_, PIL.Image.Image)



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING
from typing import Any

from bentoml.exceptions import BentoMLException

if TYPE_CHECKING:
    from fastapi import FastAPI


def make_fastapi_class_views(cls: type[Any], app: FastAPI) -> None:
    from fastapi import Depends
    from fastapi.routing import APIRoute
    from fastapi.routing import APIRouter
    from fastapi.routing import APIWebSocketRoute

    from .service import get_current_service

    class_methods = [meth for meth in vars(cls).values() if callable(meth)]
    api_routes = [
        route
        for route in app.routes
        if isinstance(route, (APIRoute, APIWebSocketRoute))
        and route.endpoint in class_methods
    ]
    if not api_routes:
        return
    # Modify these routes and mount it to a new APIRouter.
    # We need to to this (instead of modifying in place) because we want to use
    # the app.include_router to re-run the dependency analysis for each routes.
    new_router = APIRouter()
    for route in api_routes:
        app.routes.remove(route)
        # Get the original endpoint and add a `Depends` to the `self` parameter
        signature = inspect.signature(route.endpoint)
        parameters = list(signature.parameters.values())
        if not parameters:
            raise BentoMLException(
                "FastAPI class views must have at least one parameter"
            )
        self_param = parameters[0].replace(default=Depends(get_current_service))
        new_params = [self_param] + [
            # Convert all parameters to keyword-only since the first param
            # is no longer positional
            param.replace(kind=inspect.Parameter.KEYWORD_ONLY)
            for param in parameters[1:]
        ]
        new_signature = signature.replace(parameters=new_params)
        setattr(route.endpoint, "__signature__", new_signature)
        new_router.routes.append(route)
    app.include_router(new_router)



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass
from functools import lru_cache
from http import HTTPStatus
from typing import TYPE_CHECKING

from ..._internal.utils.filesystem import resolve_user_filepath
from ...exceptions import InvalidArgument
from ._import_hook import LATEST_PROTOCOL_VERSION
from ._import_hook import import_generated_stubs
from ._import_hook import import_grpc

if TYPE_CHECKING:
    from enum import Enum

    import grpc

    from ..._internal.io_descriptors import IODescriptor
    from ...exceptions import BentoMLException
    from ..types import BentoServicerContext
    from ..types import ProtoField
    from ..types import RpcMethodHandler
    from ..v1 import service_pb2 as pb

else:
    pb, _ = import_generated_stubs()
    grpc, _ = import_grpc()

__all__ = [
    "grpc_status_code",
    "parse_method_name",
    "to_http_status",
    "GRPC_CONTENT_TYPE",
    "import_generated_stubs",
    "import_grpc",
    "validate_proto_fields",
    "LATEST_PROTOCOL_VERSION",
    "load_from_file",
]

logger = logging.getLogger(__name__)

# content-type is always application/grpc
GRPC_CONTENT_TYPE = "application/grpc"


def load_from_file(p: str) -> bytes:
    rp = resolve_user_filepath(p, ctx=None)
    with open(rp, "rb") as f:
        return f.read()


def validate_proto_fields(
    field: str | None, io_: IODescriptor[t.Any]
) -> str | ProtoField:
    if field is None:
        raise InvalidArgument('"field" cannot be empty.')
    accepted_fields = io_.proto_fields + ("serialized_bytes",)
    if field not in accepted_fields:
        raise InvalidArgument(
            f"'{io_.__class__.__name__}' accepts one of the following fields: '{','.join(accepted_fields)}' got '{field}' instead.",
        ) from None
    return field


@lru_cache(maxsize=1)
def http_status_to_grpc_status_map() -> dict[Enum, grpc.StatusCode]:
    # Maps HTTP status code to grpc.StatusCode
    from http import HTTPStatus

    return {
        HTTPStatus.OK: grpc.StatusCode.OK,
        HTTPStatus.UNAUTHORIZED: grpc.StatusCode.UNAUTHENTICATED,
        HTTPStatus.FORBIDDEN: grpc.StatusCode.PERMISSION_DENIED,
        HTTPStatus.NOT_FOUND: grpc.StatusCode.UNIMPLEMENTED,
        HTTPStatus.TOO_MANY_REQUESTS: grpc.StatusCode.UNAVAILABLE,
        HTTPStatus.BAD_GATEWAY: grpc.StatusCode.UNAVAILABLE,
        HTTPStatus.SERVICE_UNAVAILABLE: grpc.StatusCode.UNAVAILABLE,
        HTTPStatus.GATEWAY_TIMEOUT: grpc.StatusCode.DEADLINE_EXCEEDED,
        HTTPStatus.BAD_REQUEST: grpc.StatusCode.INVALID_ARGUMENT,
        HTTPStatus.INTERNAL_SERVER_ERROR: grpc.StatusCode.INTERNAL,
        HTTPStatus.UNPROCESSABLE_ENTITY: grpc.StatusCode.FAILED_PRECONDITION,
    }


@lru_cache(maxsize=1)
def grpc_status_to_http_status_map() -> dict[grpc.StatusCode, Enum]:
    return {v: k for k, v in http_status_to_grpc_status_map().items()}


def grpc_status_code(err: BentoMLException) -> grpc.StatusCode:
    """
    Convert BentoMLException.error_code to grpc.StatusCode.
    """
    return http_status_to_grpc_status_map().get(err.error_code, grpc.StatusCode.UNKNOWN)


def to_http_status(status_code: grpc.StatusCode) -> int:
    """
    Convert grpc.StatusCode to HTTPStatus.
    """
    status = grpc_status_to_http_status_map().get(
        status_code, HTTPStatus.INTERNAL_SERVER_ERROR
    )

    return status.value


@dataclass
class MethodName:
    """
    Represents a gRPC method name.

    Attributes:
        package: This is defined by `package foo.bar`, designation in the protocol buffer definition
        service: service name in protocol buffer definition (eg: service SearchService { ... })
        method: method name
    """

    package: str = ""
    service: str = ""
    method: str = ""

    @property
    def fully_qualified_service(self):
        """return the service name prefixed with package"""
        return f"{self.package}.{self.service}" if self.package else self.service


def parse_method_name(method_name: str) -> tuple[MethodName, bool]:
    """
    Infers the grpc service and method name from the handler_call_details.
    e.g. /package.ServiceName/MethodName
    """
    method = method_name.split("/", maxsplit=2)
    # sanity check for method.
    if len(method) != 3:
        return MethodName(), False
    _, package_service, method = method
    *packages, service = package_service.rsplit(".", maxsplit=1)
    package = packages[0] if packages else ""
    return MethodName(package, service, method), True


def wrap_rpc_handler(
    wrapper: t.Callable[
        ...,
        t.Callable[
            [pb.Request, BentoServicerContext],
            t.Coroutine[t.Any, t.Any, pb.Response | t.Awaitable[pb.Response]],
        ],
    ],
    handler: RpcMethodHandler | None,
) -> RpcMethodHandler | None:
    if not handler:
        return None
    if not handler.request_streaming and not handler.response_streaming:
        assert handler.unary_unary
        return handler._replace(unary_unary=wrapper(handler.unary_unary))
    elif not handler.request_streaming and handler.response_streaming:
        assert handler.unary_stream
        return handler._replace(unary_stream=wrapper(handler.unary_stream))
    elif handler.request_streaming and not handler.response_streaming:
        assert handler.stream_unary
        return handler._replace(stream_unary=wrapper(handler.stream_unary))
    elif handler.request_streaming and handler.response_streaming:
        assert handler.stream_stream
        return handler._replace(stream_stream=wrapper(handler.stream_stream))
    else:
        raise RuntimeError(f"RPC method handler {handler} does not exist.") from None



"""
BentoML
=======

BentoML is a Python library for building online serving systems optimized for AI apps
and model inference. It supports serving any model format/runtime and custom Python
code, offering the key primitives for serving optimizations, task queues, batching,
multi-model chains, distributed orchestration, and multi-GPU serving.

Docs: http://docs.bentoml.com
Source Code: https://github.com/bentoml/BentoML
Developer Community: https://l.bentoml.com/join-slack
Twitter/X: https://x.com/bentomlai
Blog: https://www.bentoml.com/blog
"""

from typing import TYPE_CHECKING
from typing import Any

from ._internal.configuration import BENTOML_VERSION as __version__

MODULE_ATTRS = {
    "Field": "pydantic:Field",
    # BentoML built-in types
    "load_config": "._internal.configuration:load_config",
    "save_config": "._internal.configuration:save_config",
    "set_serialization_strategy": "._internal.configuration:set_serialization_strategy",
    "Bento": "._internal.bento:Bento",
    "BentoCloudClient": "._internal.cloud:BentoCloudClient",
    "Context": "._internal.context:ServiceContext",
    "server_context": "._internal.context:server_context",
    "Model": "._internal.models:Model",
    "monitor": "._internal.monitoring:monitor",
    "Resource": "._internal.resource:Resource",
    "Runnable": "._internal.runner:Runnable",
    "Runner": "._internal.runner:Runner",
    "Strategy": "._internal.runner.strategy:Strategy",
    "Service": "._internal.service:Service",
    "Tag": "._internal.tag:Tag",
    "load": "._internal.service.loader:load",
    "Cookie": "._internal.utils.http:Cookie",
    # Bento management APIs
    "get": ".bentos:get",
    "build": ".bentos:build",
    "delete": ".bentos:delete",
    "export_bento": ".bentos:export_bento",
    "import_bento": ".bentos:import_bento",
    "list": ".bentos:list",
    "pull": ".bentos:pull",
    "push": ".bentos:push",
    "serve": ".bentos:serve",
    # Legacy APIs
    "HTTPServer": ".server:HTTPServer",
    "GrpcServer": ".server:GrpcServer",
    # New SDK
    "service": "_bentoml_sdk:service",
    "runner_service": "_bentoml_sdk:runner_service",
    "api": "_bentoml_sdk:api",
    "task": "_bentoml_sdk:task",
    "depends": "_bentoml_sdk:depends",
    "on_shutdown": "_bentoml_sdk:on_shutdown",
    "on_startup": "_bentoml_sdk:on_startup",
    "on_deployment": "_bentoml_sdk:on_deployment",
    "asgi_app": "_bentoml_sdk:asgi_app",
    "mount_asgi_app": "_bentoml_sdk:mount_asgi_app",
    "get_current_service": "_bentoml_sdk:get_current_service",
    "IODescriptor": "_bentoml_sdk:IODescriptor",
    "importing": "_bentoml_impl.loader:importing",
    "SyncHTTPClient": "_bentoml_impl.client:SyncHTTPClient",
    "AsyncHTTPClient": "_bentoml_impl.client:AsyncHTTPClient",
}


if TYPE_CHECKING:
    # Framework specific modules
    from pydantic import Field

    from _bentoml_impl.frameworks import catboost
    from _bentoml_impl.frameworks import lightgbm
    from _bentoml_impl.frameworks import mlflow
    from _bentoml_impl.frameworks import sklearn
    from _bentoml_impl.frameworks import xgboost

    from . import bentos

    # BentoML built-in types
    from ._internal.bento import Bento
    from ._internal.cloud import BentoCloudClient
    from ._internal.configuration import load_config
    from ._internal.configuration import save_config
    from ._internal.configuration import set_serialization_strategy
    from ._internal.context import ServiceContext as Context
    from ._internal.context import server_context
    from ._internal.models import Model
    from ._internal.monitoring import monitor
    from ._internal.resource import Resource
    from ._internal.runner import Runnable
    from ._internal.runner import Runner
    from ._internal.runner.strategy import Strategy
    from ._internal.service import Service
    from ._internal.service.loader import load
    from ._internal.tag import Tag
    from ._internal.utils.http import Cookie

    # Bento management APIs
    from .bentos import build
    from .bentos import delete
    from .bentos import export_bento
    from .bentos import get
    from .bentos import import_bento
    from .bentos import list
    from .bentos import pull
    from .bentos import push
    from .bentos import serve

    try:  # needs bentoml-unsloth package
        from _bentoml_impl.frameworks import unsloth
    except ModuleNotFoundError:
        pass

    from . import diffusers_simple
    from . import gradio
    from . import ray
    from . import triton
    from ._internal.frameworks import detectron
    from ._internal.frameworks import diffusers
    from ._internal.frameworks import easyocr
    from ._internal.frameworks import fastai
    from ._internal.frameworks import flax
    from ._internal.frameworks import keras
    from ._internal.frameworks import onnx
    from ._internal.frameworks import picklable_model
    from ._internal.frameworks import pytorch
    from ._internal.frameworks import pytorch_lightning
    from ._internal.frameworks import tensorflow
    from ._internal.frameworks import torchscript
    from ._internal.frameworks import transformers

    # isort: off
    from . import io
    from . import models
    from . import metrics  # Prometheus metrics client
    from . import container  # Container API
    from . import client  # Client API
    from . import batch  # Batch API
    from . import exceptions  # BentoML exceptions
    from . import monitoring  # Monitoring API
    from . import cloud  # Cloud API
    from . import deployment  # deployment API
    from . import validators  # validators

    # isort: on
    from _bentoml_impl.client import AsyncHTTPClient
    from _bentoml_impl.client import SyncHTTPClient
    from _bentoml_impl.loader import importing
    from _bentoml_sdk import IODescriptor
    from _bentoml_sdk import api
    from _bentoml_sdk import asgi_app
    from _bentoml_sdk import depends
    from _bentoml_sdk import get_current_service
    from _bentoml_sdk import images
    from _bentoml_sdk import mount_asgi_app
    from _bentoml_sdk import on_deployment
    from _bentoml_sdk import on_shutdown
    from _bentoml_sdk import on_startup
    from _bentoml_sdk import runner_service
    from _bentoml_sdk import service
    from _bentoml_sdk import task
else:
    from _bentoml_impl.frameworks.importer import FrameworkImporter

    from ._internal.utils.lazy_loader import LazyLoader as _LazyLoader

    FrameworkImporter.install()

    bentos = _LazyLoader("bentoml.bentos", globals(), "bentoml.bentos")

    # ML Frameworks
    catboost = _LazyLoader(
        "bentoml.catboost", globals(), "_bentoml_impl.frameworks.catboost"
    )
    sklearn = _LazyLoader(
        "bentoml.sklearn", globals(), "_bentoml_impl.frameworks.sklearn"
    )
    xgboost = _LazyLoader(
        "bentoml.xgboost", globals(), "_bentoml_impl.frameworks.xgboost"
    )
    lightgbm = _LazyLoader(
        "bentoml.lightgbm", globals(), "_bentoml_impl.frameworks.lightgbm"
    )
    unsloth = _LazyLoader(
        "bentoml.unsloth", globals(), "_bentoml_impl.frameworks.unsloth"
    )
    mlflow = _LazyLoader("bentoml.mlflow", globals(), "_bentoml_impl.frameworks.mlflow")
    detectron = _LazyLoader(
        "bentoml.detectron",
        globals(),
        "bentoml._internal.frameworks.detectron",
        warning="`bentoml.detectron` is deprecated since v1.4 and will be removed in a future version.",
    )
    diffusers = _LazyLoader(
        "bentoml.diffusers",
        globals(),
        "bentoml._internal.frameworks.diffusers",
        warning="`bentoml.diffusers` is deprecated since v1.4 and will be removed in a future version.",
    )
    diffusers_simple = _LazyLoader(
        "bentoml.diffusers_simple",
        globals(),
        "bentoml.diffusers_simple",
        warning="`bentoml.diffusers_simple` is deprecated since v1.4 and will be removed in a future version.",
    )
    easyocr = _LazyLoader(
        "bentoml.easyocr",
        globals(),
        "bentoml._internal.frameworks.easyocr",
        warning="`bentoml.easyocr` is deprecated since v1.4 and will be removed in a future version.",
    )
    flax = _LazyLoader(
        "bentoml.flax",
        globals(),
        "bentoml._internal.frameworks.flax",
        warning="`bentoml.flax` is deprecated since v1.4 and will be removed in a future version.",
    )
    fastai = _LazyLoader(
        "bentoml.fastai",
        globals(),
        "bentoml._internal.frameworks.fastai",
        warning="`bentoml.fastai` is deprecated since v1.4 and will be removed in a future version.",
    )

    onnx = _LazyLoader(
        "bentoml.onnx",
        globals(),
        "bentoml._internal.frameworks.onnx",
        warning="`bentoml.onnx` is deprecated since v1.4 and will be removed in a future version.",
    )
    keras = _LazyLoader(
        "bentoml.keras",
        globals(),
        "bentoml._internal.frameworks.keras",
        warning="`bentoml.keras` is deprecated since v1.4 and will be removed in a future version.",
    )
    pytorch = _LazyLoader(
        "bentoml.pytorch",
        globals(),
        "bentoml._internal.frameworks.pytorch",
        warning="`bentoml.pytorch` is deprecated since v1.4 and will be removed in a future version.",
    )
    pytorch_lightning = _LazyLoader(
        "bentoml.pytorch_lightning",
        globals(),
        "bentoml._internal.frameworks.pytorch_lightning",
        warning="`bentoml.pytorch_lightning` is deprecated since v1.4 and will be removed in a future version.",
    )
    picklable_model = _LazyLoader(
        "bentoml.picklable_model",
        globals(),
        "bentoml._internal.frameworks.picklable_model",
        warning="`bentoml.picklable_model` is deprecated since v1.4 and will be removed in a future version.",
    )
    tensorflow = _LazyLoader(
        "bentoml.tensorflow",
        globals(),
        "bentoml._internal.frameworks.tensorflow",
        warning="`bentoml.tensorflow` is deprecated since v1.4 and will be removed in a future version.",
    )
    torchscript = _LazyLoader(
        "bentoml.torchscript",
        globals(),
        "bentoml._internal.frameworks.torchscript",
        warning="`bentoml.torchscript` is deprecated since v1.4 and will be removed in a future version.",
    )
    transformers = _LazyLoader(
        "bentoml.transformers",
        globals(),
        "bentoml._internal.frameworks.transformers",
        warning="`bentoml.transformers` is deprecated since v1.4 and will be removed in a future version.",
    )

    # Integrations
    triton = _LazyLoader("bentoml.triton", globals(), "bentoml.triton")
    ray = _LazyLoader("bentoml.ray", globals(), "bentoml.ray")
    gradio = _LazyLoader("bentoml.gradio", globals(), "bentoml.gradio")

    io = _LazyLoader("bentoml.io", globals(), "bentoml.io")
    batch = _LazyLoader("bentoml.batch", globals(), "bentoml.batch")
    models = _LazyLoader("bentoml.models", globals(), "bentoml.models")
    metrics = _LazyLoader("bentoml.metrics", globals(), "bentoml.metrics")
    container = _LazyLoader("bentoml.container", globals(), "bentoml.container")
    images = _LazyLoader("bentoml.images", globals(), "bentoml.images")
    client = _LazyLoader("bentoml.client", globals(), "bentoml.client")
    server = _LazyLoader("bentoml.server", globals(), "bentoml.server")
    exceptions = _LazyLoader("bentoml.exceptions", globals(), "bentoml.exceptions")
    monitoring = _LazyLoader("bentoml.monitoring", globals(), "bentoml.monitoring")
    cloud = _LazyLoader("bentoml.cloud", globals(), "bentoml.cloud")
    deployment = _LazyLoader("bentoml.deployment", globals(), "bentoml.deployment")
    validators = _LazyLoader("bentoml.validators", globals(), "bentoml.validators")
    del _LazyLoader, FrameworkImporter

    def __getattr__(name: str) -> Any:
        if name in MODULE_ATTRS:
            from importlib import import_module

            module_name, attr_name = MODULE_ATTRS[name].split(":")
            module = import_module(module_name, __package__)
            return getattr(module, attr_name)
        raise AttributeError(f"module {__name__} has no attribute {name}")


__all__ = [
    "__version__",
    "Context",
    "Cookie",
    "Service",
    "bentos",
    "models",
    "batch",
    "metrics",
    "container",
    "server_context",
    "client",
    "io",
    "Tag",
    "Model",
    "Runner",
    "Runnable",
    "monitoring",
    "BentoCloudClient",  # BentoCloud REST API Client
    # bento APIs
    "list",
    "get",
    "delete",
    "import_bento",
    "export_bento",
    "build",
    "load",
    "push",
    "pull",
    "serve",
    "Bento",
    "exceptions",
    # Framework specific modules
    "catboost",
    "detectron",
    "unsloth",
    "diffusers",
    "diffusers_simple",
    "easyocr",
    "flax",
    "fastai",
    "lightgbm",
    "mlflow",
    "onnx",
    "picklable_model",
    "pytorch",
    "pytorch_lightning",
    "keras",
    "sklearn",
    "tensorflow",
    "torchscript",
    "transformers",
    "xgboost",
    # integrations
    "ray",
    "gradio",
    "cloud",
    "deployment",
    "triton",
    "monitor",
    "load_config",
    "save_config",
    "set_serialization_strategy",
    "Strategy",
    "Resource",
    # new SDK
    "service",
    "runner_service",
    "api",
    "task",
    "images",
    "on_shutdown",
    "on_startup",
    "on_deployment",
    "depends",
    "IODescriptor",
    "validators",
    "Field",
    "get_current_service",
    "asgi_app",
    "mount_asgi_app",
    # new implementation
    "SyncHTTPClient",
    "AsyncHTTPClient",
    "importing",
]



from __future__ import annotations

import logging
import os
import typing as t
from types import ModuleType
from typing import TYPE_CHECKING

import attr

import bentoml
from bentoml import Tag
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException
from bentoml.exceptions import NotFound
from bentoml.models import ModelContext
from bentoml.models import ModelOptions as BaseModelOptions

from ..utils.pkg import PackageNotFoundError
from ..utils.pkg import get_pkg_version

if TYPE_CHECKING:
    from bentoml.types import ModelSignature
    from bentoml.types import ModelSignatureDict

    from .utils.onnx import ONNXArgCastedType

    ProvidersType = list[str | tuple[str, dict[str, t.Any]]]


try:
    import onnx
    from google.protobuf.json_format import MessageToDict

    from .utils.onnx import gen_input_casting_func
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "onnx is required in order to use module 'bentoml.onnx', install onnx with 'pip install onnx'. For more information, refer to https://onnx.ai/get-started.html"
    )

try:
    import onnxruntime as ort
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'onnxruntime' or 'onnxruntime-gpu' is required by 'bentoml.onnx', install onnxruntime or onnxruntime-gpu with 'pip install onnxruntime' or 'pip install onnxruntime-gpu'. For more information, refer to https://onnxruntime.ai/"
    )

MODULE_NAME = "bentoml.onnx"
MODEL_FILENAME = "saved_model.onnx"
API_VERSION = "v2"

logger = logging.getLogger(__name__)


def flatten_providers_list(lst: ProvidersType) -> list[str]:
    return [k[0] if isinstance(k, (list, tuple)) else k for k in lst]


@attr.define
class ModelOptions(BaseModelOptions):
    """Options for the ONNX model"""

    input_specs: dict[str, list[dict[str, t.Any]]] = attr.field(factory=dict)
    output_specs: dict[str, list[dict[str, t.Any]]] = attr.field(factory=dict)
    providers: ProvidersType = attr.field(default=None)
    session_options: t.Optional["ort.SessionOptions"] = attr.field(default=None)


def get(tag_like: str | Tag) -> bentoml.Model:
    """
    Get the BentoML model with the given tag.

    Args:
        tag_like: The tag of the model to retrieve from the model store.

    Returns:
        :obj:`~bentoml.Model`: A BentoML :obj:`~bentoml.Model` with the matching tag.

    Example:

    .. code-block:: python

       import bentoml
       # target model must be from the BentoML model store
       model = bentoml.onnx.get("onnx_resnet50")
    """
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, not loading with {MODULE_NAME}."
        )
    return model


def _load_raw_model(bento_model: str | Tag | bentoml.Model) -> onnx.ModelProto:
    if not isinstance(bento_model, bentoml.Model):
        bento_model = get(bento_model)

    model_path = bento_model.path_of(MODEL_FILENAME)
    raw_model = onnx.load(model_path)
    return raw_model


def load_model(
    bento_model: str | Tag | bentoml.Model,
    *,
    providers: ProvidersType | None = None,
    session_options: ort.SessionOptions | None = None,
) -> ort.InferenceSession:
    """
    Load the onnx model with the given tag from the local BentoML model store.

    Args:
        bento_model (``str`` ``|`` :obj:`~bentoml.Tag` ``|`` :obj:`~bentoml.Model`):
            Either the tag of the model to get from the store, or a BentoML `~bentoml.Model`
            instance to load the model from.
        providers (`List[Union[str, Tuple[str, Dict[str, Any]]`, `optional`, default to :code:`None`):
            Different providers provided by users. By default BentoML will use ``["CPUExecutionProvider"]``
            when loading a model.
        session_options (`onnxruntime.SessionOptions`, `optional`, default to :code:`None`):
            SessionOptions per use case. If not specified, then default to :code:`None`.

    Returns:
        :obj:`onnxruntime.InferenceSession`:
            An instance of ONNX Runtime inference session created using ONNX model loaded from the model store.

    Example:

    .. code-block:: python

        import bentoml
        sess = bentoml.onnx.load_model("my_onnx_model")
    """  # noqa

    if not isinstance(bento_model, bentoml.Model):
        bento_model = get(bento_model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, not loading with {MODULE_NAME}."
        )

    if providers:
        if not all(
            i in ort.get_all_providers() for i in flatten_providers_list(providers)
        ):
            raise BentoMLException(f"'{providers}' cannot be parsed by `onnxruntime`")
    else:
        providers = ["CPUExecutionProvider"]

    return ort.InferenceSession(
        bento_model.path_of(MODEL_FILENAME),
        sess_options=session_options,
        providers=providers,
    )


def save_model(
    name: Tag | str,
    model: onnx.ModelProto,
    *,
    signatures: dict[str, ModelSignatureDict] | dict[str, ModelSignature] | None = None,
    labels: dict[str, str] | None = None,
    custom_objects: dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: dict[str, t.Any] | None = None,
) -> bentoml.Model:
    """Save a onnx model instance to the BentoML model store.

    Args:
        name (``str``):
            The name to give to the model in the BentoML store. This must be a valid
            :obj:`~bentoml.Tag` name.
        model (:obj:`~onnx.ModelProto`):
            The ONNX model to be saved.
        signatures (``dict[str, ModelSignatureDict]``, optional):
            Signatures of predict methods to be used. If not provided, the signatures default to
            ``{"run": {"batchable": False}}``. See :obj:`~bentoml.types.ModelSignature` for more details.
            ``bentoml.onnx`` internally use ``onnxruntime.InferenceSession`` to run inference.
            When the original model is converted to ONNX format and loaded by
            ``onnxruntime.InferenceSession``, the inference method of the original model is
            converted to the ``run`` method of the ``onnxruntime.InferenceSession``.
            ``signatures`` here refers to the predict method of
            ``onnxruntime.InferenceSession``, hence the only allowed method name in ``signatures``
            is ``run``.
        labels (``dict[str, str]``, optional):
            A default set of management labels to be associated with the model. An example is
            ``{"training-set": "data-1"}``.
        custom_objects (``dict[str, Any]``, optional):
            Custom objects to be saved with the model. An example is
            ``{"my-normalizer": normalizer}``.

            Custom objects are currently serialized with cloudpickle, but this implementation is
            subject to change.
        external_modules (:code:`List[ModuleType]`, `optional`, default to :code:`None`):
            user-defined additional python modules to be saved alongside the model or custom objects,
            e.g. a tokenizer module, preprocessor module, model configuration module
        metadata (``dict[str, Any]``, optional):
            Metadata to be associated with the model. An example is ``{"bias": 4}``.

            Metadata is intended for display in a model management UI and therefore must be a
            default Python type, such as ``str`` or ``int``.

    Returns:

        :obj:`~bentoml.Model`: A BentoML model containing the saved
        ONNX model instance.  store.

    Example:

    .. code-block:: python

        import bentoml

        import torch
        import torch.nn as nn

        class ExtendedModel(nn.Module):
            def __init__(self, D_in, H, D_out):
                # In the constructor we instantiate two nn.Linear modules and assign them as
                #  member variables.
                super(ExtendedModel, self).__init__()
                self.linear1 = nn.Linear(D_in, H)
                self.linear2 = nn.Linear(H, D_out)

            def forward(self, x, bias):
                # In the forward function we accept a Tensor of input data and an optional bias
                h_relu = self.linear1(x).clamp(min=0)
                y_pred = self.linear2(h_relu)
                return y_pred + bias


        N, D_in, H, D_out = 64, 1000, 100, 1
        x = torch.randn(N, D_in)
        model = ExtendedModel(D_in, H, D_out)

        input_names = ["x", "bias"]
        output_names = ["output1"]

        tmpdir = "/tmp/model"
        model_path = os.path.join(tmpdir, "test_torch.onnx")
        torch.onnx.export(
            model,
            (x, torch.Tensor([1.0])),
            model_path,
            input_names=input_names,
            output_names=output_names,
        )

        bento_model = bentoml.onnx.save_model("onnx_model", model_path, signatures={"run": {"batchable": True}})

    """

    # prefer "onnxruntime-gpu" over "onnxruntime" for framework versioning
    _onnxruntime_version = None
    _onnxruntime_pkg = None
    _PACKAGE = ["onnxruntime-gpu", "onnxruntime", "onnxruntime-silicon"]
    for p in _PACKAGE:
        try:
            _onnxruntime_version = get_pkg_version(p)
            _onnxruntime_pkg = p
            break
        except PackageNotFoundError:
            pass
    assert _onnxruntime_pkg is not None and _onnxruntime_version is not None, (
        "Failed to find onnxruntime package version."
    )

    assert _onnxruntime_version is not None, "onnxruntime is not installed"
    if not isinstance(model, onnx.ModelProto):
        raise TypeError(f"Given model ({model}) is not a onnx.ModelProto.")

    context = ModelContext(
        framework_name="onnx",
        framework_versions={
            "onnx": get_pkg_version("onnx"),
            _onnxruntime_pkg: _onnxruntime_version,
        },
    )

    if signatures is None:
        signatures = {
            "run": {"batchable": False},
        }
        logger.info(
            'Using the default model signature for ONNX (%s) for model "%s".',
            signatures,
            name,
        )
    else:
        provided_methods = list(signatures.keys())
        if provided_methods != ["run"]:
            raise BentoMLException(
                f"Provided method names {[m for m in provided_methods if m != 'run']} are invalid. 'bentoml.onnx' will load ONNX model into an 'onnxruntime.InferenceSession' for inference, so the only supported method name is 'run'."
            )

    run_input_specs = [MessageToDict(inp) for inp in model.graph.input]
    run_output_specs = [MessageToDict(out) for out in model.graph.output]
    input_specs = {"run": run_input_specs}
    output_specs = {"run": run_output_specs}

    options = ModelOptions(input_specs=input_specs, output_specs=output_specs)

    with bentoml.models._create(  # type: ignore
        name,
        module=MODULE_NAME,
        api_version=API_VERSION,
        signatures=signatures,
        labels=labels,
        options=options,
        custom_objects=custom_objects,
        external_modules=external_modules,
        metadata=metadata,
        context=context,
    ) as bento_model:
        onnx.save(model, bento_model.path_of(MODEL_FILENAME))

        return bento_model


def get_runnable(bento_model: bentoml.Model) -> t.Type[bentoml.Runnable]:
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    # backward compatibility for v1, load raw model to infer
    # input_specs/output_specs for onnx model
    if bento_model.info.api_version == "v1":
        raw_model: onnx.ModelProto | None = None
        options = t.cast(ModelOptions, bento_model.info.options)

        if not options.input_specs:
            raw_model = _load_raw_model(bento_model)
            run_input_specs = [MessageToDict(inp) for inp in raw_model.graph.input]
            input_specs = {"run": run_input_specs}
            bento_model = bento_model.with_options(input_specs=input_specs)

        if not options.output_specs:
            raw_model = raw_model or _load_raw_model(bento_model)
            run_output_specs = [MessageToDict(out) for out in raw_model.graph.output]
            output_specs = {"run": run_output_specs}
            bento_model = bento_model.with_options(output_specs=output_specs)

    class ONNXRunnable(bentoml.Runnable):
        SUPPORTED_RESOURCES = ("nvidia.com/gpu", "cpu")
        SUPPORTS_CPU_MULTI_THREADING = True

        def __init__(self):
            super().__init__()

            session_options = (
                bento_model.info.options.session_options or ort.SessionOptions()
            )

            # check for resources
            available_gpus = os.getenv("CUDA_VISIBLE_DEVICES")
            if available_gpus is not None and available_gpus not in ("", "-1"):
                # assign GPU resources
                providers = bento_model.info.options.providers or [
                    "CUDAExecutionProvider",
                    "CPUExecutionProvider",
                ]

            else:
                # assign CPU resources

                # If onnxruntime-gpu is installed,
                # CUDAExecutionProvider etc. will be available even no
                # GPU is presented in system, which may result some
                # error when initializing ort.InferenceSession
                providers = bento_model.info.options.providers or [
                    "CPUExecutionProvider"
                ]

                # set CPUExecutionProvider parallelization options
                # TODO @larme: follow onnxruntime issue 11668 and
                # 10330 to decide best cpu parallelization strategy
                thread_count = int(os.getenv("BENTOML_NUM_THREAD", 1))
                session_options.execution_mode = ort.ExecutionMode.ORT_PARALLEL
                if session_options.intra_op_num_threads != 0:
                    logger.warning(
                        "Overriding specified 'session_options.intra_op_num_threads'."
                    )
                session_options.intra_op_num_threads = thread_count
                if session_options.inter_op_num_threads != 0:
                    logger.warning(
                        "Overriding specified 'session_options.inter_op_num_threads'."
                    )
                session_options.inter_op_num_threads = thread_count

            self.model = load_model(
                bento_model, session_options=session_options, providers=providers
            )

            self.predict_fns: dict[str, t.Callable[..., t.Any]] = {}
            for method_name in bento_model.info.signatures:
                self.predict_fns[method_name] = getattr(self.model, method_name)

    def add_runnable_method(
        method_name: str,
        signatures: ModelSignature,
        input_specs: list[dict[str, t.Any]],
        output_specs: list[dict[str, t.Any]],
    ):
        casting_funcs = [gen_input_casting_func(spec) for spec in input_specs]

        if len(output_specs) > 1:

            def _process_output(outs):
                return tuple(outs)

        else:

            def _process_output(outs):
                return outs[0]

        def _run(self: ONNXRunnable, *args: t.Any) -> t.Any:
            casted_args = [
                casting_funcs[idx](args[idx]) for idx in range(len(casting_funcs))
            ]

            input_names: dict[str, ONNXArgCastedType] = {
                i.name: val for i, val in zip(self.model.get_inputs(), casted_args)
            }
            output_names: list[str] = [o.name for o in self.model.get_outputs()]
            raw_outs = self.predict_fns[method_name](output_names, input_names)
            return _process_output(raw_outs)

        ONNXRunnable.add_method(
            _run,
            name=method_name,
            batchable=signatures.batchable,
            batch_dim=signatures.batch_dim,
            input_spec=signatures.input_spec,
            output_spec=signatures.output_spec,
        )

    for method_name, signatures in bento_model.info.signatures.items():
        options = t.cast(ModelOptions, bento_model.info.options)
        input_specs = options.input_specs[method_name]
        output_specs = options.output_specs[method_name]
        add_runnable_method(method_name, signatures, input_specs, output_specs)

    return ONNXRunnable



from __future__ import annotations

import importlib.metadata
import importlib.util
import subprocess
from functools import lru_cache
from importlib.metadata import PackageNotFoundError
from pathlib import Path
from types import ModuleType
from typing import cast

from packaging.version import Version

__all__ = [
    "PackageNotFoundError",
    "pkg_version_info",
    "get_pkg_version",
    "source_locations",
    "find_spec",
]

get_pkg_version = importlib.metadata.version
find_spec = importlib.util.find_spec


@lru_cache(maxsize=None)
def pkg_version_info(pkg_name: str | ModuleType) -> tuple[int, int, int]:
    if isinstance(pkg_name, ModuleType):
        pkg_name = pkg_name.__name__
    pkg_version = Version(get_pkg_version(pkg_name))
    return pkg_version.major, pkg_version.minor, pkg_version.micro


def source_locations(pkg: str) -> str | None:
    module = find_spec(pkg)
    if module is None:
        return
    (module_path,) = module.submodule_search_locations  # type: ignore (unfinished typed)
    return cast(str, module_path)


DEFAULT_BENTOML_GIT_URL = "https://github.com/bentoml/bentoml.git"


@lru_cache(maxsize=1)
def get_local_bentoml_dependency() -> str:
    import logging

    logger = logging.getLogger(__name__)
    logger.info(
        "Getting local bentoml dependency, please make sure all local changes are pushed"
    )
    src_dir = Path(__file__).parent.parent.parent.parent.parent
    try:
        branch_name = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            cwd=src_dir,
            text=True,
        ).strip()
    except subprocess.CalledProcessError:
        branch_name = "main"
    # Precise checkout only returns "HEAD" as branch name
    if branch_name == "HEAD":
        ref = subprocess.check_output(
            ["git", "rev-parse", "HEAD"],
            cwd=src_dir,
            text=True,
        ).strip()
        remote = "origin"
    else:
        ref = branch_name
        try:
            remote = subprocess.check_output(
                ["git", "config", "--get", f"branch.{branch_name}.remote"],
                cwd=src_dir,
                text=True,
            ).strip()
        except subprocess.CalledProcessError:
            remote = "origin"

    try:
        remote_url = subprocess.check_output(
            ["git", "config", "--get", f"remote.{remote}.url"],
            cwd=src_dir,
            text=True,
        ).strip()
    except subprocess.CalledProcessError:
        remote_url = DEFAULT_BENTOML_GIT_URL

    return f"git+{remote_url}@{ref}"



from __future__ import annotations

import contextlib
import importlib
import logging
import os
import sys
import types
import typing as t
from typing import TYPE_CHECKING

import fs
import fs.mirror
from simple_di import Provide
from simple_di import inject

from ...exceptions import InvalidArgument
from ..configuration.containers import BentoMLContainer
from ..utils.cattr import bentoml_cattr
from .base import OCIBuilder
from .generate import generate_containerfile

if TYPE_CHECKING:
    from ..bento import Bento
    from ..bento import BentoStore
    from ..tag import Tag
    from .base import Arguments

    P = t.ParamSpec("P")

    class DefaultBackendImpl(types.ModuleType):
        BUILD_CMD: list[str] | None
        ENV: dict[str, str] | None
        BUILDKIT_SUPPORT: bool

        def find_binary(self) -> str | None: ...

        def construct_build_args(
            self,
            **kwargs: t.Any,  # pylint: disable=unused-argument
        ) -> Arguments: ...

        def health(self) -> bool: ...

    DefaultBuilder: t.TypeAlias = t.Literal[
        "docker", "podman", "buildah", "buildx", "nerdctl", "buildctl"
    ]


logger = logging.getLogger(__name__)

BUILDER_REGISTRY: dict[str, OCIBuilder] = {}


DEFAULT_BACKENDS = frozenset(
    {"docker", "buildx", "buildah", "podman", "nerdctl", "buildctl"}
)


def register_default_backends():
    for backend in DEFAULT_BACKENDS:
        try:
            module = t.cast(
                "DefaultBackendImpl", importlib.import_module(f".{backend}", __name__)
            )
            register_backend(
                backend,
                health=module.health,
                binary=module.find_binary(),
                construct_build_args=module.construct_build_args,
                buildkit_support=module.BUILDKIT_SUPPORT,
                env=getattr(module, "ENV", None),
                build_cmd=getattr(module, "BUILD_CMD", None),
            )
        except NotImplementedError:
            logger.debug("%s is not yet implemented.", backend)
        except Exception as err:  # pylint: disable=broad-except
            raise ValueError(f"Failed to register backend '{backend}: {err}'") from None


@inject
def determine_container_tag(
    bento_tag: Tag | str,
    image_tag: tuple[str] | None = None,
    _bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
):
    # NOTE: for tags strategy, we will always generate a default tag from the bento:tag
    # If '-t/--image-tag' is provided, we will use this tag provided by user.
    bento = _bento_store.get(bento_tag)
    tag = (str(bento.tag),)
    if image_tag is not None:
        assert isinstance(image_tag, tuple)
        tag = image_tag
    return tag


def enable_buildkit(
    *, backend: str | None = None, builder: OCIBuilder | None = None
) -> bool:
    # We will look for DOCKER_BUILDKIT in the environment variable.
    # as this will be our entry point to enable BuildKit.
    if "DOCKER_BUILDKIT" in os.environ:
        return bool(int(os.environ["DOCKER_BUILDKIT"]))
    # If the variable is not set, fallback to the default value
    # provided by the builder.
    if builder is not None:
        return builder.enable_buildkit
    elif backend is not None:
        return get_backend(backend).enable_buildkit
    else:
        raise ValueError("Either backend or builder must be provided.")


# XXX: Sync with BentoML extra dependencies found in pyproject.toml
FEATURES = frozenset(
    {
        "tracing",
        "grpc",
        "grpc-reflection",
        "grpc-channelz",
        "monitor-otlp",
        "triton",
        "aws",
        "all",
        "io",
        "io-file",
        "io-image",
        "io-pandas",
        "io-json",
        "tracing-zipkin",
        "tracing-jaeger",
        "tracing-otlp",
    }
)


@contextlib.contextmanager
def construct_containerfile(
    bento: Bento,
    enable_buildkit: bool = True,
    *,
    features: t.Sequence[str] | None = None,
    add_header: bool = False,
) -> t.Generator[tuple[str, str], None, None]:
    from _bentoml_sdk.models import BentoModel
    from _bentoml_sdk.models import HuggingFaceModel

    from ..bento.bento import BaseBentoInfo
    from ..bento.bento import BentoInfo
    from ..bento.bento import BentoInfoV2
    from ..bento.build_config import DockerOptions

    dockerfile_path = "env/docker/Dockerfile"
    instruction: list[str] = []

    with fs.open_fs("temp://") as temp_fs:
        tempdir = temp_fs.getsyspath("/")
        with open(bento.path_of("bento.yaml"), "rb") as bento_yaml:
            options = BaseBentoInfo.from_yaml_file(bento_yaml)
        # tmpdir is our new build context.
        fs.mirror.mirror(bento._fs, temp_fs, copy_if_newer=True)

        # copy models from model store
        bento_model_dir = temp_fs.makedir("models", recreate=True)
        hf_model_dir = temp_fs.makedir("hf-models", recreate=True)
        for model in options.all_models:
            if model.registry == "huggingface":
                model_ref = HuggingFaceModel.from_info(model)
                model_ref.resolve(hf_model_dir)
            else:
                model_ref = BentoModel.from_info(model)
                model_ref.resolve(bento_model_dir)

        # NOTE: dockerfile_template is already included in the
        # Dockerfile inside bento, and it is not relevant to
        # construct_containerfile. Hence it is safe to set it to None here.
        # See https://github.com/bentoml/BentoML/issues/3399.
        if isinstance(options, BentoInfo):
            docker_attrs = bentoml_cattr.unstructure(options.docker)
            if (
                "dockerfile_template" in docker_attrs
                and docker_attrs["dockerfile_template"] is not None
            ):
                # NOTE: if users specify a dockerfile_template, we will
                # save it to /env/docker/Dockerfile.template. This is necessary
                # for the reconstruction of the Dockerfile.
                docker_attrs["dockerfile_template"] = "env/docker/Dockerfile.template"

            dockerfile = generate_containerfile(
                docker=DockerOptions(**docker_attrs),
                build_ctx=tempdir,
                conda=options.conda,
                bento_fs=temp_fs,
                enable_buildkit=enable_buildkit,
                add_header=add_header,
            )
            instruction.append(dockerfile)
            if features is not None:
                diff = set(features).difference(FEATURES)
                if len(diff) > 0:
                    raise InvalidArgument(
                        f"Available features are: {FEATURES}. Invalid fields from provided: {diff}"
                    )
                PIP_CACHE_MOUNT = (
                    "--mount=type=cache,target=/root/.cache/pip "
                    if enable_buildkit
                    else ""
                )
                instruction.append(
                    "RUN %spip install bentoml[%s]"
                    % (PIP_CACHE_MOUNT, ",".join(features))
                )
        else:
            from _bentoml_impl.docker import generate_dockerfile

            assert isinstance(options, BentoInfoV2)
            dockerfile = generate_dockerfile(
                options.image,
                temp_fs,
                enable_buildkit=enable_buildkit,
                add_header=add_header,
                envs=options.envs,
            )
            instruction.append(dockerfile)
        temp_fs.writetext(dockerfile_path, "\n".join(instruction))
        yield tempdir, temp_fs.getsyspath(dockerfile_path)


@inject
def build(
    bento_tag: Tag | str,
    backend: str,
    features: t.Sequence[str] | None = None,
    _bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
    **kwargs: t.Any,
):
    clean_context = contextlib.ExitStack()

    bento = _bento_store.get(bento_tag)

    builder = get_backend(backend)
    context_path, dockerfile = clean_context.enter_context(
        construct_containerfile(
            bento,
            features=features,
            enable_buildkit=enable_buildkit(builder=builder),
        )
    )
    try:
        kwargs.update({"file": dockerfile, "context_path": context_path})
        return builder.build(**kwargs)
    except Exception as e:  # pylint: disable=broad-except
        logger.error(
            "\nEncountered exception while trying to building image: %s",
            e,
            exc_info=sys.exc_info(),
        )
    finally:
        clean_context.close()


def register_backend(
    backend: str,
    *,
    buildkit_support: bool,
    health: t.Callable[[], bool],
    construct_build_args: t.Callable[P, Arguments],
    binary: str | None = None,
    build_cmd: t.Sequence[str] | None = None,
    env: dict[str, str] | None = None,
):
    """
    Register a custom backend, provided with a build and health check implementation.

    Args:
        backend: Name of the backend.
        buildkit_support: Whether the backend has support for BuildKit.
        health: Health check implementation. This is a callable that takes no
                argument and returns a boolean.
        construct_build_args: This is a callable that takes possible backend keyword arguments
                              and returns a list of command line arguments.
        env: Default environment variables dict for this OCI builder implementation.
        binary: Optional binary path. If not provided, the binary will use the backend name.
                Make sure that the binary is on your ``PATH``.
        build_cmd: Optional build command. If not provided, the command will be 'build'.

    Examples:

    .. code-block:: python

        from bentoml.container import register_backend

        register_backend(
            "lima",
            binary=shutil.which("limactl"),
            buildkit_support=True,
            health=buildx_health,
            construct_build_args=buildx_build_args,
            env={"DOCKER_BUILDKIT": "1"},
        )
    """
    if backend in BUILDER_REGISTRY:
        raise ValueError(f"Backend {backend} already registered.")
    if binary is None:
        binary = backend
    try:
        BUILDER_REGISTRY[backend] = OCIBuilder.create(
            binary,
            env=env,
            build_cmd=build_cmd,
            enable_buildkit=buildkit_support,
            construct_build_args=construct_build_args,
            health=health,
        )
    except Exception as e:  # pylint: disable=broad-except
        logger.error(
            "Failed to register backend %s: %s", backend, e, exc_info=sys.exc_info()
        )
        raise e from None


@t.overload
def health(backend: DefaultBuilder) -> bool: ...


@t.overload
def health(backend: str) -> bool: ...


def health(backend: str) -> bool:
    """
    Check if the backend is healthy.

    Args:
        backend: The name of the backend.

    .. note::

        If given backend is a type of OCIBuilder, and the backend is not registered,
        make sure to register it with ``bentoml.container.register_backend``.

    Returns:
        True if the backend is healthy, False otherwise.
    """
    return get_backend(backend).health()


@t.overload
def get_backend(backend: DefaultBuilder) -> OCIBuilder: ...


@t.overload
def get_backend(backend: str) -> OCIBuilder: ...


def get_backend(backend: str) -> OCIBuilder:
    """
    Get a given backend.

    Raises:
        ``ValueError``: If given backend is not available in backend registry.
    """
    if isinstance(backend, OCIBuilder):
        if backend not in BUILDER_REGISTRY.values():
            logger.warning(
                "Backend '%s' not registered. To use with 'bentoml.container.build', make sure to regsiter it with 'bentoml.container.register_backend'",
                backend,
            )
        return backend
    if backend not in BUILDER_REGISTRY:
        raise ValueError(
            f"Backend {backend} not registered. Available backends: {REGISTERED_BACKENDS}."
        )
    return BUILDER_REGISTRY[backend]


register_default_backends()

REGISTERED_BACKENDS = list(BUILDER_REGISTRY.keys())

__all__ = [
    "build",
    "health",
    "register_backend",
    "get_backend",
    "REGISTERED_BACKENDS",
]



from __future__ import annotations

import collections
import datetime
import logging
import logging.config
import typing as t
from pathlib import Path

import yaml

from ..context import server_context
from ..context import trace_context
from .base import MonitorBase

if t.TYPE_CHECKING:
    from ..types import JSONSerializable


DEFAULT_CONFIG_YAML = """
version: 1
disable_existing_loggers: false
loggers:
  bentoml_monitor_data:
    level: INFO
    handlers: [bentoml_monitor_data]
    propagate: false
  bentoml_monitor_schema:
    level: INFO
    handlers: [bentoml_monitor_schema]
    propagate: false
handlers:
  bentoml_monitor_data:
    class: logging.handlers.TimedRotatingFileHandler
    level: INFO
    formatter: bentoml_json
    filename: '{data_filename}'
    when: "D"
  bentoml_monitor_schema:
    class: logging.handlers.RotatingFileHandler
    level: INFO
    formatter: bentoml_json
    filename: '{schema_filename}'
formatters:
  bentoml_json:
    class: pythonjsonlogger.jsonlogger.JsonFormatter
    format: "()"
    validate: false
"""


class DefaultMonitor(MonitorBase["JSONSerializable"]):
    """
    The default monitor implementation. It uses a logger to log data and schema, and will
    write monitor data to rotating files. The schema is logged as a JSON object, and the
    data is logged as a JSON array.
    """

    PRESERVED_COLUMNS = (COLUMN_TIME, COLUMN_RID, COLUMN_TID) = (
        "timestamp",
        "request_id",
        "trace_id",
    )

    def __init__(
        self,
        name: str,
        log_path: str,
        log_config_file: str | None = None,
        **_: t.Any,
    ) -> None:
        super().__init__(name, **_)
        self.log_config_file = log_config_file
        self.log_path = log_path
        self.data_logger = None
        self.schema_logger = None

    def _init_logger(self) -> None:
        if self.log_config_file is None:
            logging_config_yaml = DEFAULT_CONFIG_YAML
        else:
            with open(self.log_config_file, "r", encoding="utf8") as f:
                logging_config_yaml = f.read()

        worker_id = server_context.worker_index or 0
        schema_path = Path(self.log_path).joinpath(
            self.name, "schema", f"schema.{worker_id}.log"
        )
        data_path = Path(self.log_path).joinpath(
            self.name, "data", f"data.{worker_id}.log"
        )

        schema_path.parent.mkdir(parents=True, exist_ok=True)
        data_path.parent.mkdir(parents=True, exist_ok=True)

        logging_config_yaml = logging_config_yaml.format(
            schema_filename=str(schema_path.absolute()),
            data_filename=str(data_path.absolute()),
            worker_id=worker_id,
            monitor_name=self.name,
        )

        try:
            logging_config = yaml.safe_load(logging_config_yaml)
        except yaml.YAMLError as e:
            raise ValueError(
                f"Error loading logging config from {self.log_config_file}: {e}"
            ) from e

        logging.config.dictConfig(logging_config)
        self.data_logger = logging.getLogger("bentoml_monitor_data")
        self.schema_logger = logging.getLogger("bentoml_monitor_schema")

    def export_schema(self, columns_schema: dict[str, dict[str, str]]) -> None:
        """
        Export columns_schema of the data. This method should be called after all data is logged.
        """
        if self.schema_logger is None:
            self._init_logger()
            assert self.schema_logger is not None

        self.schema_logger.info(
            dict(
                meta_data={
                    "bento_name": server_context.bento_name,
                    "bento_version": server_context.bento_version,
                },
                columns=list(columns_schema.values()),
            )
        )

    def export_data(
        self,
        datas: dict[str, collections.deque[JSONSerializable]],
    ) -> None:
        """
        Export data. This method should be called after all data is logged.
        """
        if self.data_logger is None:
            self._init_logger()
            assert self.data_logger is not None

        extra_columns = {
            self.COLUMN_TIME: datetime.datetime.now().isoformat(),
            self.COLUMN_RID: str(trace_context.request_id),
            self.COLUMN_TID: str(trace_context.trace_id),
        }
        while True:
            try:
                record = {k: v.popleft() for k, v in datas.items()}
                record.update(extra_columns)
                self.data_logger.info(record)
            except IndexError:
                break



from __future__ import annotations

from starlette.applications import Starlette
from starlette.routing import Match
from starlette.routing import Mount
from starlette.types import Scope


class PassiveMount(Mount):
    """A subclass of mount that doesn't match the path prefix eagerly."""

    def matches(self, scope: Scope) -> tuple[Match, Scope]:
        match, child_scope = super().matches(scope)
        if match == Match.FULL and isinstance(self.app, Starlette):
            scope = {**scope, **child_scope}
            partial_match: Match | None = None
            for route in self.app.routes:
                child_match, _ = route.matches(scope)
                if child_match == Match.FULL:
                    return child_match, child_scope
                if child_match == Match.PARTIAL and partial_match is None:
                    partial_match = child_match

            if partial_match is not None:
                return partial_match, child_scope
            return Match.NONE, child_scope
        return match, child_scope



from __future__ import annotations

import asyncio
import inspect
import logging
import os
import sys
import typing as t
from concurrent.futures import ThreadPoolExecutor
from functools import cached_property
from functools import partial
from typing import TYPE_CHECKING

from simple_di import Provide
from simple_di import inject

from ...grpc.utils import LATEST_PROTOCOL_VERSION
from ...grpc.utils import import_generated_stubs
from ...grpc.utils import import_grpc
from ...grpc.utils import load_from_file
from ..configuration.containers import BentoMLContainer
from ..context import ServiceContext as Context
from ..utils import with_app_arg
from ..utils.lazy_loader import LazyLoader

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import grpc
    from grpc import aio
    from grpc_health.v1 import health
    from grpc_health.v1 import health_pb2 as pb_health
    from grpc_health.v1 import health_pb2_grpc as services_health
    from starlette.applications import Starlette

    from ...grpc.types import Interceptors
    from ..service import Service
    from ..types import LifecycleHook
else:
    grpc, aio = import_grpc()
    health_exception_msg = "'grpcio-health-checking' is required for using health checking endpoints. Install with 'pip install grpcio-health-checking'."
    pb_health = LazyLoader(
        "pb_health",
        globals(),
        "grpc_health.v1.health_pb2",
        exc_msg=health_exception_msg,
    )
    services_health = LazyLoader(
        "services_health",
        globals(),
        "grpc_health.v1.health_pb2_grpc",
        exc_msg=health_exception_msg,
    )
    health = LazyLoader(
        "health",
        globals(),
        "grpc_health.v1.health",
        exc_msg="'grpcio-health-checking' is required for using health checking endpoints. Install with 'pip install grpcio-health-checking'.",
    )


# NOTE: we are using the internal aio._server.Server (which is initialized with aio.server)
class Server(aio._server.Server):
    """An async implementation of a gRPC server."""

    @inject
    def __init__(
        self,
        bento_service: Service,
        bind_address: str,
        max_message_length: int | None = Provide[
            BentoMLContainer.grpc.max_message_length
        ],
        maximum_concurrent_rpcs: int | None = Provide[
            BentoMLContainer.grpc.maximum_concurrent_rpcs
        ],
        enable_reflection: bool = False,
        enable_channelz: bool = False,
        max_concurrent_streams: int | None = None,
        migration_thread_pool_workers: int = 1,
        ssl_certfile: str | None = None,
        ssl_keyfile: str | None = None,
        ssl_ca_certs: str | None = None,
        graceful_shutdown_timeout: float | None = None,
        compression: grpc.Compression | None = None,
        protocol_version: str = LATEST_PROTOCOL_VERSION,
    ):
        pb, _ = import_generated_stubs(protocol_version)

        self.bento_service = bento_service
        self.servicer = bento_service.get_grpc_servicer(protocol_version)

        # options
        self.max_message_length = max_message_length
        self.max_concurrent_streams = max_concurrent_streams
        self.bind_address = bind_address
        self.enable_reflection = enable_reflection
        self.enable_channelz = enable_channelz
        self.graceful_shutdown_timeout = graceful_shutdown_timeout
        self.ssl_certfile = ssl_certfile
        self.ssl_keyfile = ssl_keyfile
        self.ssl_ca_certs = ssl_ca_certs
        self.protocol_version = protocol_version

        # Create a health check servicer. We use the non-blocking implementation
        # to avoid thread starvation.
        self.health_servicer = health.aio.HealthServicer()

        self.mount_servicers = self.bento_service.mount_servicers

        self.service_names = tuple(
            service.full_name for service in pb.DESCRIPTOR.services_by_name.values()
        ) + (health.SERVICE_NAME,)

        super().__init__(
            # Note that the max_workers are used inside ThreadPoolExecutor.
            # This ThreadPoolExecutor are used by aio.Server() to execute non-AsyncIO RPC handlers.
            # Setting it to 1 makes it thread-safe for sync APIs.
            thread_pool=ThreadPoolExecutor(max_workers=migration_thread_pool_workers),
            generic_handlers=() if self.handlers is None else self.handlers,
            interceptors=list(map(lambda x: x(), self.interceptors)),
            options=self.options,
            # maximum_concurrent_rpcs defines the maximum number of concurrent RPCs this server
            # will service before returning RESOURCE_EXHAUSTED status.
            # Set to None will indicate no limit.
            maximum_concurrent_rpcs=maximum_concurrent_rpcs,
            compression=compression,
        )

    @inject
    async def wait_for_runner_ready(
        self,
        _: Starlette,
        *,
        check_interval: int = Provide[
            BentoMLContainer.api_server_config.runner_probe.period
        ],
    ):
        if BentoMLContainer.api_server_config.runner_probe.enabled.get():
            logger.info("Waiting for runners to be ready...")
            logger.debug("Current runners: %r", self.bento_service.runners)

            while True:
                try:
                    runner_statuses = (
                        runner.runner_handle_is_ready()
                        for runner in self.bento_service.runners
                    )
                    runners_ready = all(await asyncio.gather(*runner_statuses))

                    if runners_ready:
                        break
                except ConnectionError as e:
                    logger.debug("[%s] Retrying ...", e)

                await asyncio.sleep(check_interval)

            logger.info("All runners ready.")

    @property
    def options(self) -> grpc.aio.ChannelArgumentType:
        options: grpc.aio.ChannelArgumentType = []

        if sys.platform != "win32":
            # https://github.com/grpc/grpc/blob/master/include/grpc/impl/codegen/grpc_types.h#L294
            # Eventhough GRPC_ARG_ALLOW_REUSEPORT is set to 1 by default, we want still
            # want to explicitly set it to 1 so that we can spawn multiple gRPC servers in
            # production settings.
            options.append(("grpc.so_reuseport", 1))
        if self.max_concurrent_streams:
            options.append(("grpc.max_concurrent_streams", self.max_concurrent_streams))
        if self.enable_channelz:
            options.append(("grpc.enable_channelz", 1))
        if self.max_message_length:
            options.extend(
                (
                    # grpc.max_message_length this is a deprecated options, for backward compatibility
                    ("grpc.max_message_length", self.max_message_length),
                    ("grpc.max_receive_message_length", self.max_message_length),
                    ("grpc.max_send_message_length", self.max_message_length),
                )
            )

        return tuple(options)

    @property
    def interceptors(self) -> Interceptors:
        # Note that order of interceptors is important here.

        from ...grpc.interceptors.opentelemetry import (
            AsyncOpenTelemetryServerInterceptor,
        )

        interceptors: Interceptors = [AsyncOpenTelemetryServerInterceptor]

        if BentoMLContainer.api_server_config.metrics.enabled.get():
            from ...grpc.interceptors.prometheus import PrometheusServerInterceptor

            interceptors.append(PrometheusServerInterceptor)

        if BentoMLContainer.api_server_config.logging.access.enabled.get():
            from ...grpc.interceptors.access import AccessLogServerInterceptor

            access_logger = logging.getLogger("bentoml.access")
            if access_logger.getEffectiveLevel() <= logging.INFO:
                interceptors.append(AccessLogServerInterceptor)

        # add users-defined interceptors.
        interceptors.extend(self.bento_service.interceptors)

        return interceptors

    @property
    def handlers(self) -> t.Sequence[grpc.GenericRpcHandler] | None:
        # Note that currently BentoML doesn't provide any specific
        # handlers for gRPC. If users have any specific handlers,
        # BentoML will pass it through to grpc.aio.Server
        return self.bento_service.grpc_handlers

    @cached_property
    def loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_event_loop()

    def run(self) -> None:
        try:
            self.loop.run_until_complete(self.serve())
        finally:
            try:
                self.loop.call_soon_threadsafe(
                    lambda: asyncio.ensure_future(self.shutdown())
                )
            except Exception as e:  # pylint: disable=broad-except
                raise RuntimeError(f"Server failed unexpectedly: {e}") from None

    @cached_property
    def context(self) -> Context:
        return Context()

    def configure_port(self, addr: str):
        if self.ssl_certfile:
            client_auth = False
            ca_cert = None
            assert self.ssl_keyfile, (
                "'ssl_keyfile' is required when 'ssl_certfile' is provided."
            )
            if self.ssl_ca_certs is not None:
                client_auth = True
                ca_cert = load_from_file(self.ssl_ca_certs)
            server_credentials = grpc.ssl_server_credentials(
                (
                    (
                        load_from_file(self.ssl_keyfile),
                        load_from_file(self.ssl_certfile),
                    ),
                ),
                root_certificates=ca_cert,
                require_client_auth=client_auth,
            )

            self.add_secure_port(addr, server_credentials)
        else:
            self.add_insecure_port(addr)

    async def serve(self) -> None:
        self.configure_port(self.bind_address)
        await self.startup()
        await self.wait_for_termination()

    @property
    def on_startup(self) -> list[LifecycleHook]:
        on_startup = [
            *map(with_app_arg, self.bento_service.startup_hooks),
            with_app_arg(self.bento_service.on_grpc_server_startup),
        ]
        if BentoMLContainer.development_mode.get():
            for runner in self.bento_service.runners:
                on_startup.append(with_app_arg(partial(runner.init_local, quiet=True)))
        else:
            for runner in self.bento_service.runners:
                if runner.embedded:
                    on_startup.append(
                        with_app_arg(partial(runner.init_local, quiet=True))
                    )
                else:
                    on_startup.append(with_app_arg(runner.init_client))

        on_startup.append(self.wait_for_runner_ready)
        return on_startup

    async def startup(self) -> None:
        from ...exceptions import MissingDependencyException

        _, services = import_generated_stubs(self.protocol_version)

        # Running on_startup callback.
        for handler in self.on_startup:
            out = handler(None)
            if inspect.isawaitable(out):
                await out

        # register bento servicer
        services.add_BentoServiceServicer_to_server(self.servicer, self)
        services_health.add_HealthServicer_to_server(self.health_servicer, self)

        service_names = self.service_names
        # register custom servicer
        for (
            user_servicer,
            add_servicer_fn,
            user_service_names,
        ) in self.mount_servicers:
            add_servicer_fn(user_servicer(), self)
            service_names += tuple(user_service_names)
        if self.enable_channelz:
            try:
                from grpc_channelz.v1 import channelz
            except ImportError:
                raise MissingDependencyException(
                    "'--debug' is passed, which requires 'grpcio-channelz' to be installed. Install with 'pip install bentoml[grpc-channelz]'."
                ) from None
            if "GRPC_TRACE" not in os.environ:
                logger.debug(
                    "channelz is enabled, while GRPC_TRACE is not set. No channel tracing will be recorded."
                )
            channelz.add_channelz_servicer(self)
        if self.enable_reflection:
            try:
                # reflection is required for health checking to work.
                from grpc_reflection.v1alpha import reflection
            except ImportError:
                raise MissingDependencyException(
                    "reflection is enabled, which requires 'grpcio-reflection' to be installed. Install with 'pip install bentoml[grpc-reflection]'."
                ) from None
            service_names += (reflection.SERVICE_NAME,)
            reflection.enable_server_reflection(service_names, self)
        # mark all services as healthy
        for service in service_names:
            await self.health_servicer.set(
                service,
                pb_health.HealthCheckResponse.SERVING,  # type: ignore (no types available)
            )
        await self.start()

    @property
    def on_shutdown(self) -> list[LifecycleHook]:
        on_shutdown = [
            *map(with_app_arg, self.bento_service.shutdown_hooks),
            with_app_arg(self.bento_service.on_grpc_server_shutdown),
        ]
        for runner in self.bento_service.runners:
            on_shutdown.append(with_app_arg(runner.destroy))

        return on_shutdown

    async def shutdown(self):
        # Running on_shutdown callback.
        for handler in self.on_shutdown:
            out = handler(None)
            if inspect.isawaitable(out):
                await out

        await self.stop(grace=self.graceful_shutdown_timeout)
        await self.health_servicer.enter_graceful_shutdown()
        self.loop.stop()
